"""Utilities for converting local or online CIF data to NCMAT data.

Two utility classes (CIFSource and CIFLoader) are included, and along with the
NCMATComposer they are used to implement the related cmd-line scripts (in
particular ncrystal_cif2ncmat), but they can of course also be used directly via
the Python API below by expert users.

"""

################################################################################
##                                                                            ##
##  This file is part of NCrystal (see https://mctools.github.io/ncrystal/)   ##
##                                                                            ##
##  Copyright 2015-2023 NCrystal developers                                   ##
##                                                                            ##
##  Licensed under the Apache License, Version 2.0 (the "License");           ##
##  you may not use this file except in compliance with the License.          ##
##  You may obtain a copy of the License at                                   ##
##                                                                            ##
##      http://www.apache.org/licenses/LICENSE-2.0                            ##
##                                                                            ##
##  Unless required by applicable law or agreed to in writing, software       ##
##  distributed under the License is distributed on an "AS IS" BASIS,         ##
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  ##
##  See the License for the specific language governing permissions and       ##
##  limitations under the License.                                            ##
##                                                                            ##
################################################################################

__all__ = ['CIFSource','CIFLoader', 'produce_validation_plot','produce_validation_plots']

from . import _common as _nc_common
from . import ncmat as _nc_ncmat
from . import _ncmatimpl as _nc_ncmatimpl
from . import core as _nc_core

class CIFSource:

    """
    Generalised CIF source, either in the form of text data, a file path, a
    database ID (in the form of a string like "codid::9008460" or "mpid::87)",
    for either the Crystallography Open Database
    (https://www.crystallography.net/cod/) or the Materials Project
    (https://www.materialsproject.org/), an existing CIFSource object, or any
    object like CIFLoader which has a cifsrc property which is a CIFSource
    object.
    """

    def __init__( self, data, *, allow_fail = False ):
        """
        Initialise from data in various formats (see class
        description). Unless allow_fail=True, an unrecognised input format will
        result in an NCBadInput exception being thrown. If allow_fail=True the
        .invalid property can be used to check if the loading failed.
        """
        self.__codid,self.__mpid,self.__fp,self.__textdata = None,None,None,None

        #Check if data is CIFSource object or has one as a .cifsrc property (e.g. LoadedCIF, CIFAnalyser)
        o = getattr( data, 'cifsrc', None ) or data
        if isinstance(o,CIFSource):
            self.__codid = o.__codid
            self.__mpid = o.__mpid
            self.__fp = o.__fp
            self.__textdata = o.__textdata
            return
        def _setfp( pth ):
            import pathlib
            pth = pth.decode() if hasattr(pth,'decode') else pth
            self.__fp = pathlib.Path( pth )
        if hasattr(data,'__fspath__'):
            return _setfp(data)
        if hasattr( data, 'startswith' ):
            str_to_sb = ( lambda s : s ) if not hasattr(data,'decode') else ( lambda s : s.encode() )

            if data.startswith(str_to_sb('codid::')) and data[7:].isdigit():
                self.__codid = int(data[7:])
                return
            if data.startswith(str_to_sb('mpid::')) and data[6:].isdigit():
                self.__mpid = int(data[6:])
                return
        if isinstance( data, str ):
            if '\n' in data.strip():
                self.__textdata = data
            else:
                _setfp(data)
            return
        if isinstance( data, bytes ):
            if b'\n' in data.strip():
                self.__textdata = data.decode()
            else:
                _setfp(data)
            return
        if not allow_fail:
            _nc_core.NCBadInput('Could not detect CIF source type')

    @property
    def invalid( self ):
        """Check if instance is invalid (only possible if constructed with allow_fail=True). """
        return all(e is None for e in (self.__codid,self.__mpid,self.__fp,self.__textdata))

    @property
    def codid( self ):
        """
        None or a database ID (integer) for the Crystallography Open Database
        (https://www.crystallography.net/cod/).
        """
        return self.__codid

    @property
    def mpid( self ):
        """
        None or a database ID (integer) for the Materials Project
        (https://www.materialsproject.org/).
        """
        return self.__mpid

    @property
    def textdata( self ):
        """
        None or a string containing raw CIF data.
        """
        return self.__textdata

    @property
    def filepath( self ):
        """
        None or a pathlib.Path to an on-disk CIF file.
        """
        return self.__fp

    @property
    def is_remote( self):
        """True if either .codid or .mpid is not None"""
        return any( ( not e is None ) for e in (self.__codid,self.__mpid) )

    def load_data( self, quiet = False, mp_apikey = None ):
        """Try to load the data and return it as a string. Depending on the
        source, this might result in reading a file or querying an online
        database. In case the data is defined via a materials project ID (mpid),
        a materials project API key must be provided either via the mp_apikey
        parameter, or in the MATERIALSPROJECT_USER_API_KEY environment variable.
        """
        assert not self.invalid
        if self.__textdata is not None:
            return self.__textdata
        if self.__fp is not None:
            if not quiet:
                _nc_common.print(f"Loading data from file {self.__fp.name}")
            return self.__fp.read_text()
        if self.__mpid is not None:
            return _mp_get_cifdata( self.__mpid, quiet = quiet, apikey = mp_apikey )
        assert self.__codid
        return _cod_get_cifdata( self.__codid, quiet = quiet )

class CIFLoader:

    """Class which is used to load and analyse CIF data, primarily so it can be
    used to transfer the crystal structure into an NCMATComposer object, with
    the intention of producing NCMAT data and new materials for usage with
    NCrystal.

    Behind the scenes the class uses third-party projects gemmi and spglib to
    parse the CIF data and process the symmetries within.

    The perhaps biggest issue to understand when basing NCrystal materials on
    CIF data, is that usually CIF data does not contain any or only incomplete
    information about the dynamics of the atoms. Thus, for a complete
    description of a high-quality material, one might want to pair up the usage
    of a CIFLoader, with a PhononDOSAnalyser from the NCrystal.vdos
    module.

    However, some CIF data contain information about atomic displacements
    (either "Uiso" or the anisotropic equivalent), which NCrystal is able to
    convert into a Debye temperature in order to provide a temperature-dependent
    model of dynamics and atomic displacement. However, such "Uiso" information
    is only useful if one knows at which temperature it was obtained, and that
    information is almost always absent in most CIF data out there. Thus, in the
    last step - when using the .create_ncmat(..) or .create_ncmat_composer(..)
    methods of the CIFLoader object, one should ideally use the uiso_temperature
    parameter to provide this information to NCrystal (usually after a call to
    print(cifloader.raw_cif_textdata) to manually inspect the meta-data, read
    any linked journal papers, etc.).

    TODO: Add link here to online tutorial once it is ready.

    """

    def __init__( self, cifsrc, quiet = False, mp_apikey = None,
                  refine_with_spglib = True, merge_equiv = True, override_spacegroup = None ):
        """Initialise from cifsrc which can either be a CIFSource object, or any
        sort of data which can be used to initialise a CIFSource object (cf. the
        CIFSource class documentation)

        If refine_with_spglib=False, the structure will not be verified and
        refined with spglib. It is strongly recommended to not disable this
        refinement, as it is needed to ensure correctness and consistency of the
        resulting material.

        The override_spacegroup parameter can in some cases be used to select a
        particular setting of the detected spacegroup. For instance, spacegroup
        number 70 is available in two settings, and loading a CIF file with
        those might result in a warning that both "F d d d:1" and "F d d d:2"
        are available. Trying again with override_spacegroup="F d d d:1" can be
        used to explicitly select one of them (it might require some
        investigation of course to determine which one is right). A list of
        space group settings can be found at:

        https://cci.lbl.gov/sginfo/hall_symbols.html

        Setting merge_equiv=False will prevent usage of the same NCMAT label for
        multiple lists of equivalent atomic positions. However, note that this
        might affect the symmetry and space group of the structure, and is in
        general not recommended.

        If quiet=True, no informative messages will be emitted.

        Finally, the mp_apikey can be used to specify an API key in case the
        cifsrc is an entry for the materialsproject.org. To avoid having to
        specify the key here, one can instead provide the key in the environment
        variable MATERIALSPROJECT_USER_API_KEY.
        """
        with _nc_common.WarningSpy( block = quiet ) as warnlist:
            load, gemmi, struct = _actual_init_gemmicif( cifsrc,
                                                         quiet = quiet,
                                                         mp_apikey = mp_apikey,
                                                         refine_with_spglib = refine_with_spglib,
                                                         merge_equiv = merge_equiv,
                                                         override_spacegroup = override_spacegroup )
        assert len(load)==9
        self.__cifsrc = load['cifsrc']
        self.__actual_codid = load['actual_codid']
        self.__actual_mpid = load['actual_mpid']
        self.__cif_chemformula = load['cif_chemformula']
        self.__cifdata = load['cifdata']
        self.__cif_raw = load['cif_raw']
        self.__cellsg = load['cellsg']
        self.__atoms = load['atoms']
        self.__cifdescr = load['cifdescr']
        self.__warnlist = tuple( warnlist )

        #Hidden support for additional processing in derived class:
        _ = getattr( self, '_process_raw_gemmi_objects', None )
        if _:
            _( gemmi, struct )

    @property
    def cifsrc( self ):
        """The CIFSource object on which everything is based."""
        return self.__cifsrc

    @property
    def actual_codid( self ):
        """None or the database ID of the Crystallography Open Database. This
        might be present even if .cifsrc.codid is None, since the CIF data might
        have been downloaded manually, but provide the ID in its metadata."""
        return self.__actual_codid

    @property
    def actual_mpid( self ):
        """None or the database ID of the Material Project. This might be
        present even if .cifsrc.mpid is None, since the CIF data might have
        been downloaded manually, but provide the ID in its metadata.
        """
        return self.__actual_mpid

    @property
    def raw_cif_textdata( self ):
        """The raw CIF input data, as a string."""
        return self.__cifdata

    @property
    def raw_cif_chemformula( self ):
        """The chemical formula encoded in the CIF data (usually in the
        _chemical_formula_sum field) as a string."""
        return self.__cif_chemformula

    @property
    def raw_cif_dict( self ):
        """The CIF data parsed into a dictionary."""
        return self.__cif_raw

    @property
    def atoms( self ):
        """A tuple of all atoms in the crystal structure, including positions in
        the unit cell. The format of each atom is essentially a dictionary
        similar to:
               {'aniso': None,
                'cif_labels': ['Al'],
                'composition': ((1.0, 'Al'),),
                'equivalent_positions': [(0.0, 0.0, 0.0),
                                         (0.0, 0.5, 0.5),
                                         (0.5, 0.0, 0.5),
                                         (0.5, 0.5, 0.0)],
                'occupancy': 1.0,

                'uiso': None})

        Here 'aniso' might provide information about
        anisotropic displacements, 'uiso' (Uiso) about isotropic displacements,
        and occupancy is the site occupancy. The 'cif_labels' are the
        corresponding labels in the CIF data, and the 'composition' +
        'equivalent_positions' should be self-explanatory.
        """
        return self.__atoms

    @property
    def cellsg( self ):
        """
        Get information about unit cell parameters and space group, as a
        dictionary with a format like:

           {'a': 4.0496, 'b': 4.0496,'c': 4.0496,
            'alpha': 90,'beta': 90,'gamma': 90,
            'spacegroup': {'hm': 'Fm-3m', 'number': 225} }

        """
        return self.__cellsg

    @property
    def extracted_description( self ):
        """
        Description of the material based on the meta-data found in the CIF
        data. Returned as a list of strings.
        """
        return self.__cifdescr

    @property
    def warnings( self ):
        """
        List of warnings emitted during processing of the CIF data, as a
        sequence of tuples of two strings: (warning_message, warning_category).
        """
        return self.__warnlist

    def create_ncmat( self, *args, **kwargs ):
        """
        Return NCMAT data corresponding to the currently held crystal
        parameters. All parameters are forwarded (same as calling
        create_ncmat_composer(*args,**kwargs).create_ncmat())
        """
        c = self.create_ncmat_composer(*args,**kwargs)
        return c.create_ncmat()

    def create_ncmat_composer( self, *,
                               uiso_temperature = None,
                               skip_dyninfo = False,
                               top_comments = None,
                               fallback_debye_temp = 300.0,
                               quiet = False,
                               remap = None,
                               no_formula_check = False ):
        """Setup and return an NCMATComposer object, based on the currently held
        crystal parameters.

        The most important thing to be aware of is that, if you wish to take
        advantage of any Uiso/biso values found in the CIF data, you must
        provide the uiso_temperature parameter value (kelvin), to indicate the
        temperature at which these are valid (most likely after inspecting the
        .raw_cif_textdata manually for hints about the correct value).

        If quiet=True, no informative messages will be emitted.

        The fallback_debye_temp parameter can be used to change the default
        Debye temperature assigned to atoms, for which Uiso/biso information was
        not available.

        The other parameters are more esoteric:

        If skip_dyninfo=True, no ncmat.set_dyninfo_... calls will be performed,
        presumably because the caller will perform such calls subsequently
        (i.e. with a PhononDOSAnalyser).

        The remap parameter can be used to remap elements and isotopes found in
        the CIF data to any desired composition (cf. the
        NCMATComposer.set_composition for the allowed composition syntax. For
        instance, if one wishes to take a CIF file containing hydrogen atoms and
        deuterate it partly, one could used remap = { 'H' : '0.95 D 0.05 H', }.

        Finally, unless no_formula_check=True, an exception is raised if the
        chemical formula of the resulting material is incompatible with any
        chemical formula indicated in the input CIF meta data. This check is
        normally nice to keep, since it can catch many cases where the wrong
        setting of a spacegroup was used.
        """
        return _impl_create_ncmat_composer( self,
                                            uiso_temperature = uiso_temperature,
                                            skip_dyninfo = skip_dyninfo,
                                            top_comments = top_comments,
                                            fallback_debye_temp = fallback_debye_temp,
                                            #TODO? #merge_equiv = merge_equiv,
                                            quiet = quiet,
                                            remap = remap,
                                            no_formula_check = no_formula_check )


def produce_validation_plots( files, verbose_lbls = True, pdf_target = None,
                              **plot_kwargs ):
    """Function which can produce several of the validation plots resulting
    from usage of the produce_validation_plot(..) function, and possibly even
    embed them in a PDF file. Any plot_kwargs will be passed along to the
    produce_validation_plot(..) function.
    """
    from NCrystal.plot import _import_matplotlib_plt, _import_matplotlib_pdfpages, _fakepyplot_mode

    if pdf_target:
        pdfpages = _import_matplotlib_pdfpages()
        the_real_pdfpages = ( pdfpages.the_real_inspected_object
                              if hasattr(pdfpages,'the_real_inspected_object')
                              else None )
        already_pdfpages = isinstance(pdf_target,the_real_pdfpages) if the_real_pdfpages else False
        pdf = pdf_target if already_pdfpages else pdfpages(pdf_target)
        assert pdf
    else:
        pdf, already_pdfpages = None, False

    if pdf:
        plt = _import_matplotlib_plt()
    for f in files:
        produce_validation_plot( f, verbose_lbls = verbose_lbls, do_show = not pdf,
                                 line_width_scale = 0.5, **plot_kwargs )
        if pdf:
            pdf.savefig()
            plt.close()

    if pdf and not already_pdfpages:
        pdf.close()
        _nc_common.print("created %s"%pdf_target)

def produce_validation_plot( data_or_file, verbose_lbls = True, line_width_scale = 1,
                             quiet = False, xlabel = None, legend_args = None, do_newfig = True,
                             do_show = True, do_legend = True, do_grid=True, do_tight_layout=True ):

    """Function which can produce validation plots for NCMAT data defining
    crystalline materials like the ones on
    https://github.com/mctools/ncrystal/wiki/Data-library. The plot will compare
    the powder Bragg diffraction cross section curve of the NCMAT file, with
    those obtained by replacing the crystal structure with a crystal structure
    loaded directly from online database entries that might be mentioned in the
    comments of the NCMAT data. This makes it easy to verify that the crystal
    structure is still compatible (or not) with those crystal structures
    mentioned in the comments.

    Currently two online databases are supported, with recognised URLs in a
    format like either of the following:

       https://www.crystallography.net/cod/9008460.html
       https://www.materialsproject.org/materials/mp-87

    """

    def lookupAndProduce( cifsrc, *,dynamics, remap):
        cifloader = CIFLoader( cifsrc, quiet = quiet )
        ncmat = cifloader.create_ncmat_composer( uiso_temperature = None,#dynamics provided separately here
                                                 remap = remap,
                                                 skip_dyninfo = True )
        ncmat.transfer_dyninfo_objects( dynamics )
        return ncmat.create_ncmat()

    from NCrystal.plot import _import_matplotlib_plt
    plt = _import_matplotlib_plt()
    if do_newfig:
        plt.figure()

    multcreate = lambda data : _nc_core.directMultiCreate(data,cfg_params='comp=bragg')
    _file = None
    fn = None
    if hasattr( data_or_file, '__fspath__' ):
        pass
    elif isinstance( data_or_file, _nc_core.TextData ):
        mc = multcreate( data_or_file )
        contentiterable = data_or_file
        fn = data_or_file.dataSourceName
    elif isinstance( data_or_file, bytes ) or isinstance( data_or_file, str ):
        _ = data_or_file.decode() if isinstance( data_or_file, bytes ) else data_or_file
        if '\n' in _ or _.startswith('NCMAT'):
            contentiterable = _.splitlines()
            mc = multcreate(_)
            fn = 'Anonymous NCMAT data'
    if not fn:
        fn = data_or_file
        contentiterable = _nc_core.createTextData(data_or_file)
        mc = multcreate(contentiterable)

    if hasattr(fn,'__fspath__'):
        import pathlib
        fn = pathlib.Path(fn).name
    elif '/' in fn:
        fn = fn.split('/')[-1]
    if not quiet:
        _nc_common.print(f'Attempting to validate {fn}')

    if not mc.info.isCrystalline():
        if not quiet:
            _nc_common.print(f"Ignoring non-crystalline material {fn}")
        return

    def _extractID(s,pattern):
        if not pattern in s:
            return None
        l=s.split(pattern)[1:]
        while l:
            e,l = l[0],l[1:]
            d=''
            while e and e[0].isdigit():
                d+=e[0]
                e=e[1:]
            if d and int(d)>0:
                yield int(d)

    import re as _re
    _re_atomdbspecs = _re.compile("\[ *with ([a-zA-Z ]+)->([ a-zA-Z0-9+-\.]+) *\]")
    def _extractAtomDBSpec(s):
        #Look for remapping specs like "[with H->D]" and return in @ATOMDB format
        #(i.e. "H is D").
        if not '->' in s:
            return
        m = _re_atomdbspecs.search(s)
        return ' '.join(('%s is %s'%m.groups()).split()) if m else None

    ids = []
    for l in contentiterable:
        atomdb = _extractAtomDBSpec(l)
        for e in _extractID(l,'materialsproject.org/materials/mp-'):
            ids.append( dict(dbtype='mp',entryid=e,atomdb=atomdb))
        for e in _extractID(l,'crystallography.net/cod/'):
            ids.append( dict(dbtype='cod',entryid=e,atomdb=atomdb))

    #order-preserving remove duplicates:
    _seen = set()
    newids=[]
    for d in ids:
        key = tuple(sorted((k,v) for k,v in d.items()))
        if key in _seen:
            continue
        newids.append(d)
        _seen.add(key)
    ids = newids

    def _atomdb_to_remap( atomdb):
        _atomdb = list( ' '.join(e.strip().split()) for e in (atomdb or '').replace(':',' ').split('@') )
        _atomdb = list( e for e in _atomdb if e )
        l = []
        for c in _atomdb:
            p=c.replace(':',' ').split()
            if not len(p)>=3 or p[1]!='is':
                raise _nc_core.NCBadInput('invalid atomdb remap syntax in "%s"'%c)
            l.append( (p[0],' '.join(p[2:]) ) )
        return l

    dynamics = mc.info
    cmps = [ (fn, mc ) ]
    for _ in ids:
        dbname = _['dbtype']
        entryid = _['entryid']
        atomdb = _['atomdb']
        lpargs=dict(remap=_atomdb_to_remap(atomdb),dynamics=dynamics)
        if dbname=='mp':
            lpargs['cifsrc']='mpid::%i'%entryid
        else:
            assert dbname=='cod'
            lpargs['cifsrc']='codid::%i'%entryid
        out = lookupAndProduce( **lpargs )
        lbl=f'{dbname}-{entryid}'
        if atomdb:
            lbl = f'{lbl} with replacement "{atomdb}"'
        cmps.append( ( lbl, multcreate(out) ) )
    from .constants import ekin2wl
    wlmax = max(1.05*ekin2wl(mc.scatter.domain()[0]) for _,mc in cmps)
    from ._numpy import _np_linspace
    wls = _np_linspace(0.001,wlmax,10000)
    if len(cmps)<=1:
        _nc_common.warn(f"No online DB IDs found in {fn}")
        return

    sgno_all = set()
    natoms_all = set()

    col_ordered = [_nc_common._palette_Few.get(k,k) for k in
                   ('black',
                    'blue',
                    'orange',
                    'green',
                    'red',
                    'brown',
                    'purple',
                    'yellow',
                    'pink',
                    'gray')]

    for i,(lbl,mc) in enumerate(cmps):
        lbl = str(lbl)
        if '/' in lbl:
            lbl=lbl.split('/')[-1]
        elif lbl.startswith('mp-'):
            lbl = 'Materials Project entry '+lbl[3:]
        elif lbl.startswith('cod-'):
            lbl = 'Crystallography Open Database entry '+lbl[4:]
        lw=4 if i>0 else 2
        lw *= line_width_scale
        args={}

        si = mc.info.structure_info if mc.info.hasStructureInfo() else None
        if not quiet:
            _nc_common.print('Structure[%s]:'%lbl,si)
        if si and verbose_lbls:
            sgno=si.get('spacegroup',None)
            lbl += ' (SG-%s, %i atoms/cell, Vcell=%g)'%(sgno or 'unspecified',si['n_atoms'],si['volume'])
            sgno_all.add(sgno)
            natoms_all.add(si['n_atoms'])

        plt.plot(wls,mc.scatter.xsect(wl=wls),
                 label=lbl,
                 linewidth=lw,
                 alpha=0.5 if i==0 else 0.5,
                 color = col_ordered[i] if i<len(col_ordered) else None,
                 dashes = [] if i==0 else [4 if len(cmps)>2 else 2,2]+[2,2]*i)
    if len(natoms_all)>1:
        _nc_common.warn('WARNING validation detected different number of atoms/cell in tested curves')
    if len(sgno_all)>1:
        _nc_common.warn('WARNING validation detected different space group numbers in tested curves')

    plt.xlim(0.0)
    plt.ylim(0.0)
    plt.title('NB: This compares the crystal structure (space group, lattice, atom positions). Phonons/dynamics always taken from .ncmat file',fontsize=6)
    plt.xlabel(xlabel or 'Neutron wavelength (%s)'%(b'\xc3\x85'.decode()))
    plt.ylabel('Coherent elastic cross section (barn)')
    if do_legend:
        plt.legend(loc='best',handlelength=5,**(legend_args or {}))
    if do_grid:
        plt.grid()
    if do_tight_layout:
        plt.tight_layout()
    if do_show:
        plt.show()

def _extract_descr_from_cif( raw_cif_dict, cifsrc, ciftextdata ):

    def normalise_result( _l ):
        #split any newlines and trim whitespace before returning:
        res = []
        for e in _l:
            for ee in e.splitlines():
                res.append( ee.rstrip() )
        return res

    if not raw_cif_dict:
        _nc_common.warn('Could not properly decode CIF metadata due to inability to access raw CIF data as dictionary.')
        return [], None, None, None

    #cif is cif section with publication info:
    l=[]
    def extract(key,*altkeys,expectlist = False):
        if not any( key in s for _,s in sorted(raw_cif_dict.items())):
            if altkeys:
                return extract( altkeys[0], *altkeys[1:], expectlist = expectlist )
            return [] if expectlist else ''
        s = list( d[key] for _,d in sorted(raw_cif_dict.items()) if key in d)[0]
        if not expectlist:
            if s is None:
                return ''
            if not hasattr(s,'strip') and hasattr(s,'__len__') and len(s)==1:
                s=s[0]#convert single element list to first element
            s=str(s).strip()
            return '' if s in ('?','.') else s
        else:
            _l=list( ('' if (e is None or (hasattr(e,'strip') and e.strip()=='?')) else str(e).strip()) for e in s )
            return list( e for e in _l if e )

    title = extract('_publ_section_title','_citation_title')
    authors = extract('_publ_author_name','_citation_author_name',expectlist=True)
    journalname = extract('_journal_name_full','_citation_journal_full')
    year = extract('_journal_year','_citation_journal_year','_citation_year')
    doi = extract('_journal_paper_doi')
    chemformsum = extract('_chemical_formula_sum','_cod_original_formula_sum')

    extracted_codid = extract('_cod_database_code')
    if extracted_codid:
        extracted_codid = int(extracted_codid) if extracted_codid.isdigit() else None

    extracted_mpid = None
    _sp = 'Note from NCrystal.cifutils: Data from materialsproject.org / mp-'# NB this exact format is produced elsewhere in this file.
    for e in ciftextdata.splitlines():
        if _sp in e:
            _ = e.split(_sp,1)
            if len(_) >= 2:
                _ = _[1].split()[0].split('#',1)[0]
                if _.isdigit():
                    extracted_mpid = int(_)

    if extracted_codid is not None and cifsrc.codid is not None and cifsrc.codid != extracted_codid:
        _nc_common.warn(f'Embedded CODID {extracted_codid} is different than requested {cifsrc.codid}!')

    if extracted_mpid is not None and cifsrc.mpid is not None and cifsrc.mpid != extracted_mpid:
        _nc_common.warn(f'Embedded MPID {extracted_mpid} is different than requested {cifsrc.mpid}!')

    _thecodid = cifsrc.codid or extracted_codid
    _thempid = cifsrc.mpid or extracted_mpid

    audit_creation_method = extract('_audit_creation_method')
    audit_creation_date = extract('_audit_creation_date')
    audit_author_name = extract('_audit_author_name')

    nauthors = len(authors or [])
    if authors:
        if len(authors)==1:
            authors = authors[0]
        elif len(authors)==2:
            authors = f'{authors[0]} and {authors[1]}'
        else:
            authors = f'{authors[0]}, et al.'
    if title:
        l.append(f'"{title}"')
    if  (journalname and year) and not authors:
        authors = '<unknown authors>'
    if journalname and year:
        _jy = f'{journalname}, {year}'
    elif journalname or year:
        _jy = journalname or year
    else:
        _jy = ''
    if authors:
        l.append(f'{authors} [{_jy}]' if _jy else f'Author{"" if nauthors==1 else "s"}: {authors}')

    if doi:
        l.append(f'DOI: https://dx.doi.org/{doi}')

    if audit_creation_method:
        l.append(f'CIF creation method: {audit_creation_method}')
    if audit_creation_date:
        l.append(f'CIF creation date: {audit_creation_date}')
    if audit_author_name:
        l.append(f'CIF created by: {audit_author_name}')

    if _thecodid:
        l += [ f'Crystallography Open Database entry {_thecodid}', _codid2url(_thecodid) ]
    if _thempid:
        l += [ 'The Materials Project', _mpid2url(_thempid) ]

    return normalise_result( l ), _thecodid, _thempid, chemformsum

def _codid2url( codid ):
    return f'https://www.crystallography.net/cod/{codid}.html'

def _mpid2url( mpid ):
    return f'https://www.materialsproject.org/materials/mp-{mpid}'

def _impl_create_ncmat_composer( cifloader, *,
                                 uiso_temperature,
                                 skip_dyninfo,
                                 top_comments,
                                 fallback_debye_temp,
                                 #merge_equiv,
                                 quiet,
                                 remap,
                                 no_formula_check ):
        with _nc_common.WarningSpy( block = quiet ) as extra_ana_warnings:
            composer = _impl_create_ncmat_composer_internal( cifloader = cifloader,
                                                             #merge_equiv = merge_equiv,
                                                             uiso_temperature = uiso_temperature,
                                                             skip_dyninfo = skip_dyninfo,
                                                             top_comments = top_comments,
                                                             remap = remap,
                                                             no_formula_check = no_formula_check )
            if not skip_dyninfo and fallback_debye_temp and fallback_debye_temp > 0.0:
                composer.allow_fallback_dyninfo( fallback_debye_temp )

        def fmtwarnings( warnings, descrtxt, nwmax ):
            l=[]
            if not warnings:
                return l
            l.append(f'Notice: The following WARNINGS were emitted {descrtxt}:')
            l.append('')

            wleft=warnings[::-1]
            while wleft:
                w = wleft.pop()
                ncount = 1 + wleft.count(w)
                if ncount > 1:
                    wleft = [e for e in wleft if e!=w]
                wmsg,wcat = w
                wmsg = list(e.strip() for e in wmsg.splitlines() if e.strip())
                s0 = '  %s :'%wcat if ncount==1 else '  (%ix) %s : '%(ncount,wcat)
                for i,m in enumerate(wmsg):
                    if i==nwmax:
                        break
                    pref=s0 if i==0 else ' '*(len(s0))
                    if i+1==nwmax and len(wmsg)>nwmax:
                        m='<%i lines of output hidden>'%(len(wmsg)-nwmax)
                    l.append('%s %s'%(pref,m))
            return l

        composer.add_comments( fmtwarnings( list(cifloader.warnings) + extra_ana_warnings,
                                            'when loading the CIF data',nwmax=10),
                               add_empty_line_divider = True )
        return composer

def _aniso_nontrivial( aniso_dict ):
    return aniso_dict and ( any( abs(aniso_dict[e])>1e-13 for e in ('u12','u13','u23') )
                            or ( aniso_dict['u11'] != aniso_dict['u22'] )
                            or ( aniso_dict['u11'] != aniso_dict['u33'] ) )


def _impl_create_ncmat_composer_internal( cifloader, *, uiso_temperature, skip_dyninfo, top_comments, remap, no_formula_check ):

    src = cifloader

    src_atoms = src.atoms

    #Apply remap:
    def _stdatomname( name ):
        return {'D':'H2','T':'H3'}.get(name,name)

    remap_decoded = {}
    for elemiso, compos in (remap or []):
        elemiso, compos = _nc_ncmatimpl._decode_composition(elemiso,compos)
        elemiso = _stdatomname( elemiso )
        remap_decoded[ elemiso ] = compos
    remap = remap_decoded
    remap_str = ''

    if remap:
        new_src_atoms = []
        for atom in src_atoms:
            if not any( e in remap for fr,e in atom['composition'] ):
                new_src_atoms.append( atom )
                continue
            newcompos = []
            for fr,e in atom['composition']:
                remap_compos = remap.get(e,None)
                if remap_compos is None:
                    newcompos.append( (fr,e) )
                else:
                    for remap_fr, remap_elem in remap_compos:
                        newcompos.append( ( fr*remap_fr, remap_elem ) )
                d = dict( (k,v) for k,v in atom.items() if k!='composition' )
                d['composition'] = newcompos
                new_src_atoms.append( d )
        src_atoms = tuple( new_src_atoms )
        #For the description:
        remap_strs = []
        for elemiso, compos in sorted(remap.items()):
            compos_str = compos[0][1] if len(compos)==1 else ' '.join( f'{_f:g} {_n}' for _f,_n in compos)
            remap_strs.append(f'{elemiso} -> {compos_str}')
        remap_str = ', '.join(remap_strs)

    #DYNINFO analysis can only be done now, where we know the skip_dyninfo and uiso_temperature values:
    dyninfo_args = []

    ncmat = _nc_ncmat.NCMATComposer()
    all_atompos = []
    for idx, a in enumerate( src_atoms ):
        lbl = f'cif_species_{idx}'
        compos = [ (fr,e) for fr,e in a['composition'] ]
        if len(compos)==1 and compos[0][0]==1.0:
            composstr = compos[0][1]
        else:
            composstr = ' '.join(f'{f:g} {n}' for f,n in compos)
        ncmat.set_composition( lbl, compos )
        if not skip_dyninfo:
            uiso = a.get('uiso',None)
            if uiso is not None:
                assert uiso>0.0
                if uiso_temperature is None:
                    _nc_common.warn(f'ignoring uiso info present in CIF input for "{composstr}" since uiso_temperature parameter value is not provided\n')
                else:
                    ncmat.set_dyninfo_msd( lbl, msd=uiso, temperature=uiso_temperature )
            _aniso = a.get('aniso',None)

            if _aniso_nontrivial(_aniso):
                _nc_common.warn('Anisotropic displacement for %s ignored (%s)'%(composstr,', '.join (f'{k}={v:g}' for k,v in sorted(_aniso.items()))))

        occu = a.get('occupancy',1.0)
        for c in a['equivalent_positions']:
            all_atompos.append( ( lbl, c[0], c[1], c[2], occu ) )

    ncmat.set_atompos( all_atompos )

    st = src.cellsg
    ncmat.set_cellsg( a=st['a'], b=st['b'], c=st['c'],
                      alpha=st['alpha'], beta=st['beta'], gamma=st['gamma'],
                      spacegroup=st['spacegroup']['number'] )

    if top_comments is not None:
        ncmat.add_comments(top_comments)

    _extracted_description = src.extracted_description
    if not _extracted_description:
        if src.cifsrc.filepath:
            _extracted_description = [f'CIF file: {src.cifsrc.filepath.name}']
        else:
            _extracted_description = ['Anonymous CIF data']
    if _extracted_description:
        ncmat.add_comments(['Structure converted (with NCrystal.cifutils module) from:',''])

        _thedescr = _extracted_description
        if remap_str:
            src.actual_mpid
            _codidurl = _codid2url(src.actual_codid) if src.actual_codid else None
            _mpidurl = _mpid2url(src.actual_mpid) if src.actual_mpid else None
            if _codidurl and _codidurl in _thedescr:
                _=f'{_codidurl} [with {remap_str}]'
                _thedescr = [e.replace(_codidurl,_) for e in _thedescr]
            elif _mpidurl and _mpidurl in _thedescr:
                _=f'{_mpidurl} [with {remap_str}]'
                _thedescr = [e.replace(_mpidurl,_) for e in _thedescr]
            else:
                _thedescr+=['',f'Note: With custom remappings: {remap_str}']

        ncmat.add_comments(['  '+e for e in _thedescr])
        ncmat.add_comments([''])

    ncmat.add_comments('IMPORTANT NOTICE: This is a mostly automatic conversion which has not been',add_empty_line_divider=True)
    ncmat.add_comments('                  verified!  In particular the @DYNINFO sections might need')
    ncmat.add_comments('                  post-editing. Before distributing this file to other people,')
    ncmat.add_comments('                  please review this, amend the comments here to document,')
    ncmat.add_comments('                  anything done, and remove this notice.')

   #formula
    def _extract_atomcount( atom ):
        occu, npos = atom['occupancy'], len(atom['equivalent_positions'])
        return npos * ( int(occu) if occu == int(occu) else float(occu) )
    total_composition = []
    for a in src_atoms:
        atomcount = _extract_atomcount( a )
        for elemfrac, eleminfo in a['composition']:
            total_composition.append( ( eleminfo, elemfrac * atomcount ) )
    formula = _nc_common.format_chemform( total_composition )


    expected_formula = cifloader.raw_cif_chemformula

    if expected_formula:
        def formula_to_dict( f ):
            if not f:
                return None
            if isinstance(f,str):
                return formula_to_dict( _decode_formula(f) )
            if isinstance(f,dict):
                return f
            res = {}
            for lbl,count in f:
                assert count>=0.0
                if count==0:
                    continue
                if lbl in res:
                    res[lbl] += count
                else:
                    res[lbl] = count
            return res or None
        from .atomdata import allElementNames
        _all_elements = allElementNames()
        def _decode_formula( s ):
            l1 = [e for e in _all_elements if len(e)==1]+['D','T']
            l2 = [e for e in _all_elements if len(e)==2]
            res = []
            while s.strip():
                s=s.strip()
                c = [ e for e in l2 if s.startswith(e) ]
                if not c:
                    c = [ e for e in l1 if s.startswith(e) ]
                if len(c)!=1:
                    return None
                c = c[0]
                s=s[len(c):].strip()
                leaddigits = ''
                while s and s[0].isdigit():
                    leaddigits += s[0]
                    s=s[1:]#nb: no strip here, we want to stop on a space!
                n = int(leaddigits) if leaddigits else 1
                res.append( (c,n) )
            return formula_to_dict(res)
        def formulas_incompatible( formula1, formula2 ):
            f1 = formula_to_dict(formula1)
            f2 = formula_to_dict(formula2)
            if not f1 or not f2:
                return False
            #map D,T,H2,H3 -> H to ensure fewer false positives:
            _ = lambda d : dict( (('H' if k in ('D','T','H2','H3') else k),v) for k,v in d.items())
            f1,f2 = _(f1),_(f2)
            if f1 == f2:
                return False
            if not set(f1.keys())==set(f2.keys()):
                return True
            kref = list(f1.keys())[0]
            assert f1[kref] > 0.0 and f2[kref]>0.0
            f2scale = f1[kref]/f2[kref]
            tol = 1e-2#not tighter to avoid false positives!
            for k,v1 in f1.items():
                v2 = f2[k]*f2scale
                if abs(v1-v2)>tol and abs(v1-v2)/(1e-300+abs(v1)+abs(v2)) > tol:
                    return True
            return False

        if not no_formula_check and expected_formula:
            actual_expected_formula_dict = formula_to_dict( expected_formula )
            if remap:
                newf = {}
                def _collect( _name, _frac ):
                    _name = _stdatomname(_name)
                    if not _name in newf:
                        newf[_name] = _frac
                    else:
                        newf[_name] += _frac
                for name, frac in actual_expected_formula_dict.items():
                    _remapped = remap.get(_stdatomname(name),None)
                    if _remapped is not None:
                        for remap_frac, remap_name in _remapped:
                            _collect( remap_name, remap_frac * frac )
                    else:
                        _collect( name, frac )
                actual_expected_formula_dict = newf
            if formulas_incompatible(total_composition,actual_expected_formula_dict):
                s = f'"{expected_formula}"'
                if remap:
                    _ = _nc_common.format_chemform( list( sorted(actual_expected_formula_dict.items() ) ) )
                    s += f' remapped to "{_}"'
                raise _nc_core.NCBadInput(f'Formula encoded in CIF data ({s}) is not compatible with formula of loaded structure ("{formula}")')

    return ncmat


def _impl_merge_atoms( atoms ):
    l = []
    for a in atoms:
        pos = list(a['equivalent_positions'])
        cif_labels = list( a['cif_labels'] )
        other_metadata = list( sorted( (k,v) for k,v in a.items() if not k in ('equivalent_positions','cif_labels') ) )
        found = False
        for k,v in l:
            if k == other_metadata:
                v[0] += list( pos )
                v[1] += list( cif_labels )
                found = True
                break
        if not found:
            l.append( (other_metadata,[list(pos),list(cif_labels)]) )
    res = []
    for other_metadata, ( pos, cif_labels ) in l:
        d = dict( (k,v) for k,v in sorted(other_metadata) )
        d['equivalent_positions'] = list( sorted( pos ) )
        d['cif_labels'] = list( sorted( cif_labels ) )
        res.append( d )
    return res

def _suggest_filename( ncmat_metadata, cifloader ):
    l = ['autogen']
    l.append( ncmat_metadata['chemform'] )
    sgnum = ncmat_metadata.get('cellsg',{}).get('spacegroup',None)
    if sgnum:
        l.append( 'sg%i'%sgnum )
    if cifloader.actual_codid:
        l.append( 'cod%i'%cifloader.actual_codid )
    if cifloader.actual_mpid:
        l.append( 'mp%i'%cifloader.actual_mpid )
    return '_'.join(l) + '.ncmat'

def _format_spglib_cell( cellsg, atoms ):
    lattice = _nc_ncmatimpl._cellparams_to_spglib_lattice(cellsg)
    atomic_points = []
    atomic_types = []
    for i,a in enumerate(atoms):
        for p in a['equivalent_positions']:
            atomic_points.append( p )
            atomic_types.append( i )
    return ( lattice, atomic_points, atomic_types )

def _impl_refine_cell( cellsg, atoms ):
    orig_cell = _format_spglib_cell( cellsg, atoms )

    d = _nc_ncmatimpl._spglib_refine_cell( orig_cell )#, symprec = 0.01, allow_axis_swap = True )
    assert len(d)==7

    refined_cell = d['refined_cell']
    new_atom_pos = dict( (i,[]) for i in range(max(refined_cell[2])+1) )
    for pos, atomidx in zip(refined_cell[1],refined_cell[2]):
        new_atom_pos[ atomidx ] += [ (pos[0],pos[1],pos[2]) ]

    new_atoms = [ dict( (k,v if k!='equivalent_positions' else new_atom_pos[idx])
                        for k,v in e.items()) for idx,e in enumerate(atoms) ]
    if not d['can_keep_anisotropic_properties']:
        new_atoms = [ dict( (k,v if k!='aniso' else None)
                            for k,v in e.items()) for idx,e in enumerate(new_atoms) ]




    new_cellsg = dict( d['cellparams_snapped'].items() )
    new_cellsg['spacegroup'] = dict( number = d['sgno'],
                                     hm = d['sgsymb_hm'] )

    return new_cellsg, new_atoms, d['warnings'], d['msgs']

def _gemmi_wrap_to_unit( gemmi_fractional ):
    #gemmi's wrap_to_unit() can return 1.0, but we want those to be 0.0 (so
    #(a,b,1.0) and (a,b,0.0) does not appear to be two different
    #points). Update: I am not 100% sure about this, but keeping this function
    #for added robustness.
    c = gemmi_fractional.wrap_to_unit()
    for i in range(3):
        if c[i]==1.0:
            c[i]=0.0
    return c

def _getMaterialsProjectAPIKEY():
    import os
    apikey = os.environ.get('MATERIALSPROJECT_USER_API_KEY',None)
    if len(apikey.strip()) < 25:
        raise _nc_core.NCException('ERROR: Legacy API key found in MATERIALSPROJECT_USER_API_KEY. Please instead provide a new API key from https://materialsproject.org/api (Keys for the new API are ~32 characters, whereas keys for the legacy API are ~16 characters).')
    if not apikey:
        raise _nc_core.NCException('ERROR: Missing API key for materialsproject.org access. To fix, '
                           +'please make sure the environment variable MATERIALSPROJECT_USER_API_KEY'
                           +' contains your personal access key that you see on'
                           +' https://www.materialsproject.org/dashboard (after logging in).')
    return apikey

def _use_local_cif_cache( fn, text_data = None, quiet = False ):
    #NB: This simple implementation use no locking to guard against race
    #conditions!!! But we do perform write+move instead of simply write, which
    #is a bit more "atomic".
    import os
    if os.environ.get('NCRYSTAL_ONLINEDB_FORBID_NETWORK'):
        def notfound():
            raise RuntimeError('Error: Trying to access remote DB but NCRYSTAL_ONLINEDB_FORBID_NETWORK is set')
        time_limit_hours = 24*7*365*1000#sure, bug me in 3023
    else:
        notfound = lambda : None
        time_limit_hours = 24*7
    import os
    d = os.environ.get('NCRYSTAL_ONLINEDB_CACHEDIR')
    if not d:
        return notfound()
    import os.path
    import pathlib
    p = pathlib.Path(os.path.expanduser(d)).resolve().absolute() if d else None
    if not p:
        return notfound()
    pfn = ( p / fn )
    if text_data:
        #store data into local cache:
        if not quiet:
            _nc_common.print(f"Adding {fn} to local file cache in $NCRYSTAL_ONLINEDB_CACHEDIR")
        p.mkdir(parents=True, exist_ok=True)
        pfn_tmp = p / f'{fn}_tmp_{os.getpid()}'
        pfn_tmp.write_text(text_data)
        pfn_tmp.replace(pfn)
        return notfound()
    #retrieve data from local cache:
    if not pfn.exists():
        return notfound()
    #Since online data can change, entries expire eventually:
    import datetime
    timelim = datetime.timedelta( hours = time_limit_hours )
    now = datetime.datetime.now(tz=datetime.timezone.utc)
    mtime = datetime.datetime.fromtimestamp(pfn.stat().st_mtime, tz=datetime.timezone.utc)
    if ( now-mtime ) > timelim:
        #entry expired:
        if not quiet:
            _nc_common.print(f"Ignoring (and removing) expired {fn} from local file cache in $NCRYSTAL_ONLINEDB_CACHEDIR")
        pfn.unlink()
        return notfound()
    if not quiet:
        _nc_common.print(f"Getting {fn} from local file cache in $NCRYSTAL_ONLINEDB_CACHEDIR")
    return pfn.read_text()

_cod_cache = []
def _cod_get_cifdata( codid, quiet = False ):
    for _codid, _result in _cod_cache:
        if codid == _codid:
            if not quiet:
                _nc_common.print(f"Using cached Crystallography Open Database result for entry {codid}")
            return _result
    cache_fn = 'cod_%i.cif'%codid
    #check file cache:
    c = _use_local_cif_cache( cache_fn, quiet = quiet )
    if c:
        return c
    if not quiet:
        _nc_common.print(f"Querying the Crystallography Open Database for entry {codid}")
    result = _nc_common.download_url( "https://www.crystallography.net/cod/%i.cif"%(codid) )
    if len(_cod_cache)==10:
        _cod_cache.pop(0)
    _cod_cache.append( (codid, result ) )
    #update file cache:
    _use_local_cif_cache( cache_fn, quiet = quiet, text_data = result )
    return result

_mp_cache = []
def _mp_get_cifdata( mpid, quiet = False, apikey = None ):
    if hasattr(mpid,'startswith') and mpid.startswith('mp-'):
        mpid = mpid[3:]
    if hasattr(mpid,'startswith') and mpid.startswith('mpid::'):
        mpid = mpid[6:]
    mpid = int(mpid)
    for _mpid, _result in _mp_cache:
        if mpid == _mpid:
            if not quiet:
                _nc_common.print(f"Using cached materialsproject.org result for entry mp-{mpid}")
            return _result
    #check file cache:
    cache_fn = 'mp_%i.cif'%mpid
    c = _use_local_cif_cache( cache_fn, quiet = quiet )
    if c:
        return c

    if not quiet:
        _nc_common.print(f"Querying materialsproject.org for entry mp-{mpid}")

    if apikey is None:
        apikey = _getMaterialsProjectAPIKEY()

    with _nc_common.WarningSpy(blockfct = lambda msg, cat : cat in ('PendingDeprecationWarning','DeprecationWarning') ):
        try:
            import mp_api.client
        except ImportError:
            raise ImportError('Could not import mp_api.client. Installing the mp-api package will most likely'
                              ' fix this (perhaps with a command like "conda install -c conda-forge mp-api"'
                              ' or "python3 -mpip install mp-api").')

    with _nc_common.WarningSpy(blockfct = lambda msg,cat : msg.lower().startswith('mpcontribs-client not installed') ):
        with mp_api.client.MPRester(apikey) as mpr:
            s = mpr.get_structure_by_material_id( f'mp-{mpid}', conventional_unit_cell=True )
            #If we do no use the symprec argument in the next line, we will get a an unrefined P1 structure:
            result = s.to(fmt='cif',symprec=1e-4, significant_figures=15, angle_tolerance=5.0, refine_struct=True)
            #But we want to sanity check that this refinement gives the same spacegroup result as listed in the MP database:
            mp_expected_sg_number = s.get_space_group_info()[1]
    sg_checked = False
    errmsg = f'Unable to reliably determine spacegroup when trying to retrieve structure for mp-{mpid} from materialsproject.org'
    for l in result.splitlines():
        p = l.split('#',1)[0].split()
        if p and p[0]=='_symmetry_Int_Tables_number':
            if not len(p)>=2 or not p[1].isdigit():
                raise _nc_core.NCBadInput(errmsg+f' (unexpected format of line: "{l}")')
            _sgnum = int(p[1])
            if _sgnum == mp_expected_sg_number:
                sg_checked = True
            else:
                raise _nc_core.NCBadInput(errmsg+f' (expected SG-{mp_expected_sg_number} but got SG-{_sgnum})')

    if not sg_checked:
        raise _nc_core.NCBadInput(errmsg+f' (no line with "_symmetry_Int_Tables_number" produced)')

    #Finally embed origin as a note (so we can extract it later for the ncmat
    #header comments). Keep the format below synchronised with the reader
    #elsewhere in this file!!
    result += f'\n# Note from NCrystal.cifutils: Data from materialsproject.org / mp-{mpid}\n'

    #Update cache and return:
    if len( _mp_cache ) == 10:
        _mp_cache.pop(0)
    _mp_cache.append( (mpid, result ) )

    #update file cache:
    _use_local_cif_cache( cache_fn, quiet = quiet, text_data = result )

    return result

def _codid2url( codid ):
    return f'https://www.crystallography.net/cod/{codid}.html'
def _mpid2url( mpid ):
    return f'https://www.materialsproject.org/materials/mp-{mpid}'

####################### NEW GEMMI STUFF ############################################
def _import_gemmi( *, sysexit = False ):
    try:
        import gemmi#both available on pypi and conda-forge
        import gemmi.cif
    except ImportError:
        m = ( 'Could not import gemmi modules needed to process CIF files.'
              +' The gemmi package is available on both PyPI ("python3 -mpip install'
              +' gemmi") and conda ("conda install -c conda-forge gemmi")' )
        if sysexit:
            raise SystemExit(m)
        else:
            raise ImportError(m)
    return gemmi, gemmi.cif

_guessmap = [None]
def _guess_spacegroup_name( gemmi, s ):
    def _guess_keys( s ):
        s=''.join(s.split())
        return [ s, s.replace('/',''),s.replace('-',''),s.replace('-','').replace('/','' ) ]
    def _init_guess_map():
        guess = {}
        def _ag( i, s ):
            if not s in guess:
                guess[s] = set([i])
            else:
                guess[s].add(i)
        def add_guess( i, s ):
            for k in _guess_keys(s):
                _ag( i, k )
        for i in range(1,230+1):
            sg = gemmi.find_spacegroup_by_number(i)
            add_guess( i, sg.xhm() )
            add_guess( i, sg.short_name() )
            add_guess( i, sg.hall )
            add_guess( i, sg.hm )
        return guess
    if not _guessmap[0]:
        _guessmap[0] = _init_guess_map()
    gm = _guessmap[0]
    possible_sgnos = set()
    for k in _guess_keys(s):
        for i in gm.get(k,[]):
            possible_sgnos.add( i )
    return list(sorted(possible_sgnos))

def _load_with_gemmi( cifblock, allow_fixup = True ):
    #Load cifblock into gemmi struct. We might perform in-place editing of the
    #cifblock, if gemmi does not immediately recognise the space group.
    gemmi, gemmi_cif = _import_gemmi()

    struct = gemmi.make_small_structure_from_block( cifblock )
    assert struct
    sg = struct.find_spacegroup()
    if sg or not allow_fixup:
        return struct, sg

    #Let us see if we can find the spacegroup with a bit of manual
    #intervention. If we can, we update the cifblock and do a full reload
    #(otherwise the gemmi struct object won't have the spacegroup and images
    #properly initialised):

    _ = cifblock.find(['_space_group_IT_number'])
    if not _:
        _ = cifblock.find(['_symmetry_Int_Tables_number'])

    if _ and len(_)==1 and len(_[0])==1 and str(_[0][0]).isdigit() and (1<=int(str(_[0][0]))<=230):
        _sgnum = int(_[0][0])
        #NB: We can do this for some groups only (cubic?) Or perhaps we could use it as a double-check only?
        sg = gemmi.find_spacegroup_by_number(_sgnum)
        assert sg and sg.number == _sgnum

    sg_hm = str(struct.spacegroup_hm).strip()
    if sg_hm and not sg:
        sg_searchname = lambda x : gemmi.find_spacegroup_by_name(x, struct.cell.alpha, struct.cell.gamma )
        attempts = [ sg_hm ]
        #TODO: We used to do this, but currently it does not make a difference:
        #for e in ('H','R'):#NB H+R is not enough, see table 6 at http://cci.lbl.gov/sginfo/hall_symbols.html
        #    if sg_hm.endswith(e):
        #        attempts.append( sg_hm[:-1] + ' : ' + sg_hm[-1] )
        for a in attempts:
            a = ' '.join(a.split())
            sg = sg_searchname( a )
            if sg:
                _nc_common.warn(f'Had to interpret spacegroup "{sg_hm}" as "{a}" before Gemmi could recognise it.')
                break

    if sg_hm and not sg:
        possible = _guess_spacegroup_name( gemmi, sg_hm )
        if len(possible) == 1:
            sgnum = possible[0]
            sg = gemmi.find_spacegroup_by_number(sgnum)
            _nc_common.warn(f'Had to interpret spacegroup "{sg_hm}" as "{sg.hm}" before Gemmi could recognise it.')
        elif len(possible) > 1:
            poshm = list( (sgnum,gemmi.find_spacegroup_by_number(sgnum).hm) for sgnum in possible )
            poshm = ', '.join( '"%s"(number %i)'  for sgnum,sgstr in poshm )
            _nc_common.warn(f'Failed to interpret spacegroup interpret spacegroup "{sg_hm}". Could be any of: {poshm}.')
    if not sg:
        #manual intervention did not help:
        return struct, None

    #patch up the cifblock with info about manually found spacegroup, and try again:
    _update_spacegroup_in_cifblock( cifblock, sg )

    return _load_with_gemmi( cifblock, allow_fixup = False )

def _update_spacegroup_in_cifblock_from_name( gemmi, cifblock, name ):
    sg = gemmi.find_spacegroup_by_name( name )
    if not sg:
        raise _nc_core.NCBadInput(f'Unknown spacegroup: "{name}"')
    _update_spacegroup_in_cifblock( cifblock, sg )

def _update_spacegroup_in_cifblock( cifblock, sg, use_xhm = True ):
    cifblock.set_pair('_space_group_name_H-M_alt', sg.xhm() if use_xhm else sg.hm )
    cifblock.set_pair('_space_group_IT_number', str(sg.number) )
    cifblock.set_pair('_space_group_name_Hall', sg.hall )
    cifblock.set_pair('_space_group_crystal_system', sg.crystal_system_str() )


def _actual_init_gemmicif( cifsrc, *, quiet, mp_apikey, refine_with_spglib, merge_equiv, override_spacegroup = None ):

    from math import fsum as _math_fsum

    result = {}
    pos_tolerance = 0.0001#NB: Best if value matches the one in NCInfoBuilder.cc

    from types import MappingProxyType
    dict_ro, list_ro = MappingProxyType, tuple

    gemmi, gemmi_cif = _import_gemmi()

    if not isinstance( cifsrc, CIFSource ):
        cifsrc = CIFSource( cifsrc )

    result['cifsrc'] = cifsrc

    _cifdata = cifsrc.load_data( quiet = quiet, mp_apikey = mp_apikey )
    if not quiet:
        _nc_common.print("Attempting to load CIF data with gemmi")

    result['cifdata'] = _cifdata
    try:
        cif_doc = gemmi_cif.read_string( _cifdata )
    except ValueError as e:
        errmsg = 'CIF parsing error (from Gemmi): "%s"'%e
        cif_doc = None
    if cif_doc is None:
        raise _nc_core.NCBadInput(errmsg or 'Unknown CIF parsing error from Gemmi')

    cif_doc_as_json = cif_doc.as_json()
    import json
    try:
        cif_doc_as_dict = json.loads( cif_doc_as_json )
    except json.decoder.JSONDecodeError as e:
        cif_doc_as_dict = None
        _nc_common.warn('Could not decode raw CIF data to dictionary (Gemmi bug?): JSONDecodeError("%s")'%e)

    result['cif_raw'] = cif_doc_as_dict
    if len(cif_doc) == 0:
        raise _nc_core.NCBadInput('CIF data had no blocks!')
    if len(cif_doc) == 1:
        cif_block_with_structure = cif_doc.sole_block()
    else:
        cif_block_with_structure = None
        for e in cif_doc:
            if not len(e.find(['_atom_site_fract_x'])):
                continue
            if cif_block_with_structure:
                cif_block_with_structure = None
                break
            cif_block_with_structure = e
        if not cif_block_with_structure:
            raise _nc_core.NCBadInput('Could not automatically determine block in CIF data with structure info')

    assert cif_block_with_structure

    if override_spacegroup:
        assert isinstance(override_spacegroup,str)
        _nc_common.warn(f'Overriding spacegroup to "{override_spacegroup}" due to explicit request')
        _update_spacegroup_in_cifblock_from_name( gemmi, cif_block_with_structure, override_spacegroup )

    #record original h-m-alt entry:
    orig_hm_alt = None
    _ = cif_block_with_structure.find(['_space_group_name_H-M_alt'])
    if _ and len(_)==1 and len(_[0])==1:
        orig_hm_alt = str(_[0][0]).strip()

    struct, sg = _load_with_gemmi( cif_block_with_structure )

    if not struct:
        raise _nc_core.NCBadInput('Could not load structure with Gemmi')

    if not sg:
        raise _nc_core.NCBadInput('Could not determine space group from CIF data')

    #warn if spacegroup setting might be ambiguous:
    if sg.number != 1:
        _ = list( sorted( (e.number,e.xhm(),e.is_reference_setting()) for e in gemmi.spacegroup_table() if ( e.number==sg.number and ':' in e.xhm())))
        #_ = [ (no,xhm,isref) for no,xhm,isref in _ if isref ]
        if len(_)>1:
            _str = '", "'.join( xhm for no,xhm,isref in _)
            if not orig_hm_alt or not any( (orig_hm_alt==e[1] or orig_hm_alt.replace(' ','')==e[1].replace(' ','')) for e in _ ):
                _nc_common.warn(f'SG-{sg.number} available in multiple choices ("{_str}") and'
                               ' which one was not encoded explicitly in the _space_group_name_H-M_alt CIF field. Consider overriding the space group explicitly when loading this file.')



    _ = cif_block_with_structure.find(['_atom_type_number_in_cell'])
    _ = sum((sum(([e] for e in l),[]) for l in _),[]) if _ else []
    if _ and not any( e is None for e in _ ):
        expected_tot_atom_in_orig_cell = sum(float(e) for e in _)
    else:
        expected_tot_atom_in_orig_cell = None
    if not ( isinstance(sg.number,int) or sg.number.isdigit() ) or not ( 1<=int(sg.number)<=230 ):
        raise _nc_core.NCBadInput(f'Could not determine space group from CIF data (it loaded with invalid SG number: {sg.number}')
    if not struct.cell.is_crystal():
        raise _nc_core.NCBadInput('Loaded structure is non-crystalline (according to Gemmi).')

    sgnumber = int(sg.number)

    if not struct.cell.is_compatible_with_spacegroup( sg ):
        raise _nc_core.NCBadInput('Loaded unit cell is not compatible with deduced space group (according to Gemmi).')

    cellsg = dict ( a = struct.cell.a, b = struct.cell.b, c = struct.cell.c,
                    alpha = struct.cell.alpha,
                    beta = struct.cell.beta,
                    gamma = struct.cell.gamma )

    fractcoord_approx_1angstrom = 1.0 / ( (struct.cell.a+struct.cell.b+struct.cell.c)/3.0 )


    #Although the sg object also provides sg.ccp4 and sg.hall, we record here
    #just sgnumber and xhm, since that is what spglib refinement also provides:
    cellsg['spacegroup'] = dict_ro( dict( number = sgnumber,
                                          hm = sg.xhm(),
                                         ) )

    collected_atoms = []

    def _expand_coord_to_all_other_images( coord, allow_finetune = 3 ):
        pos0 = _gemmi_wrap_to_unit(coord)
        #First find all brute-force expanded coords (ignore those within machine
        #precision of each other!):
        l = [ pos0 ]
        for candidate in ( _gemmi_wrap_to_unit(img.apply( pos0 )) for img in struct.cell.images):
            use = True
            for c in l:
                if _nc_ncmatimpl._unit_cell_point_dist(candidate,c) < 1e-10:
                    use = False
                    break
            if use:
                l.append( candidate )
        if not allow_finetune:
            return l
        #Now, check how many of these are very close to the initial point:
        lclose = [ _nc_ncmatimpl._remap_fract_pos_pt(e) for e in l if _nc_ncmatimpl._unit_cell_point_dist(pos0,e) < 0.01*fractcoord_approx_1angstrom ]
        assert len(lclose) > 0
        if len(lclose) == 1:
            #no issues, just return:
            return l
        #Input might have had inexact coordinates for atoms at special
        #positions. Fine-tune pos0 as average over the close points, and rerun:
        def _fract_delta( x1, x0 ):
            #remember that distance of (0,0,eps) and (0,0,1-eps) is 2eps, not 1-2eps.
            dx = x1 - x0
            if dx > 0.5:
                dx -= 1.0
            if dx < -0.5:
                dx += 1.0
            return dx
        def _fract_delta_pt( xyz1, xyz0 ):
            return tuple( _fract_delta( xyz1[i], xyz0[i] ) for i in range(3) )

        pos0_new = [ _math_fsum( [lclose[0][i]] + [ _fract_delta(c[i],lclose[0][i])/len(lclose) for c in lclose ] ) for i in range(3) ]
        #snap to (0,0,0):
        for i in range(3):
            if abs(pos0_new[i])< 1e-16:
                pos0_new[i] = 0.0
        fmt = lambda c : f'({c[0]:.15g},{c[1]:.15g},{c[2]:.15g})'
        _nc_common.warn('Fractional coordinate %s interpreted as special position %s to avoid numerical precision issues'%(fmt(coord),fmt(pos0_new)))
        return _expand_coord_to_all_other_images( gemmi.Fractional(pos0_new[0],pos0_new[1],pos0_new[2]), allow_finetune = (allow_finetune-1) )


    for site in struct.sites:
        if not float( site.occ ) > 0.0:
            continue
        pos0 = _gemmi_wrap_to_unit( site.fract )
        expanded_coords = _expand_coord_to_all_other_images( pos0 )

        aniso = None
        if site.aniso.nonzero:
            aniso = dict( (k, float(getattr(site.aniso,k))) for k in ('u11','u22','u33','u12','u13','u23') )
            if all( v==0.0 for v in aniso.values() ):
                aniso = None
        u_iso = float( site.u_iso )
        if not u_iso > 0.0:
            u_iso = None

        if aniso and not u_iso:
            #this logic might already exist on the gemmi-side, but to be safe we add it here as well:
            u_iso = ( aniso['u11'] + aniso['u22'] + aniso['u33'] ) / 3

        site_en = str(site.element.name).strip()
        if site_en == 'X' or not site_en:
            #Issue seen in codid::9005777, which had no atomic symbols but
            #labels like "CaM1", "CaM2", "OA1".."OA3", "OB1", "OB2", "OC1",
            #"OC3",... Correct if beginning of string starts with exactly one
            #element name, followed by either a digit or an upper case string
            #followed by digits.
            from .atomdata import allElementNames
            def guess_elem_name( site_label ):
                sl, traildigit = _nc_common._split_trailing_digit( site_label )
                if traildigit is None:
                    return
                #first check element names with 2 chars:
                candidates = [ e for e in allElementNames() if ( sl.startswith(e) and len(e)==2 ) ]
                if len(candidates) > 1:
                    return
                #Then those with 1 char if no hit already (also check for 'D' and 'T'):
                if not candidates:
                    candidates = [ e for e in allElementNames() if ( sl.startswith(e) and len(e)==1 ) ]
                    candidates += [ e for e in ('D','T') if ( sl.startswith(e) and len(e)==1 ) ]
                if len(candidates)!=1:
                    return
                c = candidates[0]
                assert site_label.startswith(c)
                leftover = site_label[len(c):].strip()
                while leftover and leftover[0].isupper()  and leftover[0].isalpha():
                    leftover = leftover[1:].strip()
                if not leftover or leftover.isdigit():
                    return c
            elem_name = guess_elem_name( site.label )
            if elem_name:
                _nc_common.warn(f'Assuming atomic symbol "{elem_name}" based on CIF label "{site.label}"')
            else:
                raise _nc_core.NCBadInput(f'Neither Gemmi nor NCrystal\'s custom code could deduce an atomic symbol for the CIF label "{site.label}"')
            elem_Z = None
        else:
            elem_name, elem_Z = site_en, int(site.element.atomic_number)

        elem_A = None
        if elem_name == 'D':
            elem_name, elem_Z, elem_A = 'H',1,2
        if elem_name == 'T':
            elem_name, elem_Z, elem_A = 'T',1,3
        elem_marker = elem_name + ( str(elem_A) if elem_A else '' )

        from .atomdata import elementNameToZValue
        z2name = elementNameToZValue( elem_name, allow_isotopes = False )
        if z2name is None:
            raise _nc_core.NCBadInput(f'Not an element name: "{elem_name}"')

        if elem_Z is None:
            elem_Z = z2name
        if elem_Z != z2name:
            raise _nc_core.NCBadInput('Wrong atomic number (%i) provided for element "%s"'%(elem_Z,elem_name))

        collected_atoms.append( dict( expanded_coords = expanded_coords,
                                      element_marker = elem_marker,
                                      occupancy = float(site.occ),
                                      cif_label = str(site.label),
                                      uiso = u_iso,
                                      aniso = aniso ) )

    #Now merge collected atoms which occupy the same sites:
    def has_same_sites( atom1, atom2 ):
        l1,l2 = atom1['expanded_coords'],atom2['expanded_coords']
        assert len(l1)>0 and len(l2)>0
        return len(l1) == len(l2) and any( ( _nc_ncmatimpl._unit_cell_point_dist(c2,l1[0])<pos_tolerance) for c2 in l2 )

    atoms_to_process, final_atoms = collected_atoms, []
    while atoms_to_process:
        atom1 = atoms_to_process[0]
        atoms_with_same_pos = [ atom1 ]
        atoms_with_different_pos = []
        for atom2 in atoms_to_process[1:]:
            if has_same_sites(atom1,atom2):
                if atom1['uiso'] != atom2['uiso'] or atom1['aniso'] != atom2['aniso']:
                    raise _nc_core.NCBadInput('atoms labelled "%s" and "%s" have same cell '%(atom1['cif_label'],atom2['cif_label'])
                                              +'positions but different uiso information. This is not supported.')
                atoms_with_same_pos.append( atom2 )
            else:
                atoms_with_different_pos.append( atom2 )

        cif_labels = list( a['cif_label'] for a in atoms_with_same_pos )
        cif_labels_str = '"%s"'%('", "'.join(cl for cl in cif_labels))

        composition = list( ( a['occupancy'], a['element_marker'] ) for a in atoms_with_same_pos )
        occupancy = _math_fsum( fr for fr, elem in composition )
        assert occupancy > 0.0
        if occupancy != 1.0:
            composition = list( (fr/occupancy, elem) for fr, elem in composition )
            if occupancy > (1.0 + 1e-6) :
                m=f'Error: Too high total occupancy ({occupancy} which is >1) for atoms with CIF labels: {cif_labels_str}'
                if abs(int(occupancy+0.5)-occupancy)<1e-10:
                    m+='. A possible cause could be multiple entries in the CIF file using atomic positions that are actually symmetrically equivalent!'
                raise _nc_core.NCBadInput(m)
            occupancy = min( 1.0, occupancy )

        final_atom_info = dict( cif_labels = list_ro( cif_labels ),
                                equivalent_positions = list_ro( (c.x,c.y,c.z) for c in atoms_with_same_pos[0]['expanded_coords'] ),
                                composition = list_ro( sorted( composition ) ),
                                occupancy = occupancy,
                                uiso = atoms_with_same_pos[0]['uiso'],
                                aniso = atoms_with_same_pos[0]['aniso'] )
        final_atoms.append( dict_ro( final_atom_info )  )
        atoms_to_process = atoms_with_different_pos

    if merge_equiv:
        final_atoms = list_ro( dict_ro(a) for a in _impl_merge_atoms( final_atoms ) )

    if expected_tot_atom_in_orig_cell:
        n_actual = sum( len(a['equivalent_positions']) for a in final_atoms )
        rd = lambda x,y : abs(x-y)/(max(1e-300,abs(x)+abs(y)))
        if rd( n_actual, expected_tot_atom_in_orig_cell ) > 0.01:
            raise _nc_core.NCBadInput(f'Expanded number of atoms per cell ({n_actual}) is different '
                                      'from what is stated explicitly in the CIF data'
                                      f' ({expected_tot_atom_in_orig_cell}). Wrong spacegroup (or'
                                      ' choice/setting improperly specified)?')

    if refine_with_spglib:
        cellsg, final_atoms, warnings, msgs = _impl_refine_cell( cellsg, final_atoms )
        for w in warnings:
            _nc_common.warn(w)
        if not quiet:
            for m in msgs:
                _nc_common.print(m)
    else:
        _nc_common.warn('Structure refinement and verification with spglib was disabled due to explicit request.')

    if int(cellsg['spacegroup']['number']) == 1:
        #The following warning might be a bit obsolete now that we usually
        #refine cells, but we keep it here for now as an extra safeguard:
        _nc_common.warn('Space-group number 1 ("P1") detected in CIF input. A lot (but not all) of CIF data'
                       ' with this space group listed is in a non-conventional format not suitable for NCrystal. Be wary!')

    result['cellsg'] = dict_ro( cellsg )
    def _atom_sort( a ):
        #well-defined sort order for reproducible output
        return ( tuple(sorted(a.get('composition',[]))), len(a.get('equivalent_positions')),a.get('uiso',None),a.get('cif_labels',None), tuple(a.items()) )
    result['atoms'] = list_ro( dict_ro(a) for a in sorted(final_atoms,key = _atom_sort) )

    result['cifdescr'], result['actual_codid'],result['actual_mpid'],result['cif_chemformula'] =_extract_descr_from_cif( result['cif_raw'], result['cifsrc'], result['cifdata'] )
    return result, gemmi, struct



"""Module with internal utilities used by several NCrystal modules"""

################################################################################
##                                                                            ##
##  This file is part of NCrystal (see https://mctools.github.io/ncrystal/)   ##
##                                                                            ##
##  Copyright 2015-2023 NCrystal developers                                   ##
##                                                                            ##
##  Licensed under the Apache License, Version 2.0 (the "License");           ##
##  you may not use this file except in compliance with the License.          ##
##  You may obtain a copy of the License at                                   ##
##                                                                            ##
##      http://www.apache.org/licenses/LICENSE-2.0                            ##
##                                                                            ##
##  Unless required by applicable law or agreed to in writing, software       ##
##  distributed under the License is distributed on an "AS IS" BASIS,         ##
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  ##
##  See the License for the specific language governing permissions and       ##
##  limitations under the License.                                            ##
##                                                                            ##
################################################################################

#Common print function for NCrystal modules (allowing one to capture output of
#NCrystal's python layer):
_stdprint = [print]

def print(*args,**kwargs):
    _stdprint[0](*args,**kwargs)

def set_ncrystal_print_fct( fct ):
    _stdprint[0] = fct
    return print

def get_ncrystal_print_fct():
    return _stdprint[0]

def warn(msg):
    """Emit NCrystalUserWarning via standard warnings.warn function"""
    from .exceptions import NCrystalUserWarning
    import warnings
    warnings.warn( NCrystalUserWarning( str(msg) ))

class WarningSpy:
    """Context manager which spies on any warnings emitted via warnings
    module and returns list of messages and categories.

    It does so by temporarily intercepting the warnings.showwarning
    function.

    Usage example:
        with WarningSpy() as warnlist:
            fct()
        print( "Warnings emitted:", warnlist )
    """

    def __init__(self,block=False,blockfct = None):
        assert not ( bool(block) and blockfct )
        self.__block = bool(block)
        self.__blockfct = blockfct
        self.__state = None

    def __enter__(self):
        #Make sure we create our own warnings.catch_warnings context
        #manager. This not only has the advantage of automatically restoring the
        #global warnings states (including the showwarning and simplefilter
        #function) upon __exit__, but also means that global warning
        #suppressions etc. will be temporarily disabled.
        self.__state = {}
        s = self.__state
        import warnings
        s['warncntxmgr'] = warnings.catch_warnings()
        s['warncntxmgr'].__enter__()
        s['orig'] = warnings.showwarning
        warnings.showwarning = self.__spy
        warnings.simplefilter("always")
        s['l'] = []
        return s['l']

    def __exit__(self,*args,**kwargs):
        assert self.__state
        import warnings
        warnings.showwarning = self.__state['orig']#not strictly needed since next line will also restore it
        self.__state['warncntxmgr'].__exit__()
        self.__state = None

    @staticmethod
    def _fmtwarning( message, category ):
        catname = None
        if category is not None:
            catname = str(getattr(category,'__name__',category)).strip()
        elif isinstance(message,Warning):
            catname = str(message.__class__.__name__).strip()
        if not catname:
            catname='Warning'
        return ( ( str(message).strip() or '<no message>'), catname )

    def __spy( self, message, category, *args, **kwargs ):
        #only store string objects, to decouple lifetimes of more complex objects:
        msg_str, cat_str = WarningSpy._fmtwarning( message, category )
        self.__state['l'].append( ( msg_str, cat_str ) )
        do_block = self.__block or ( self.__blockfct and self.__blockfct(msg_str, cat_str) )
        if not do_block:
            return self.__state['orig']( message, category, *args, **kwargs )

def find_fraction( x, tol = 1e-15, max_denom = 1000000 ):
    import fractions
    f = fractions.Fraction.from_float( x )
    f = f.limit_denominator(max_denom)
    if hasattr(f,'as_integer_ratio'):
        #py 3.8+
        a,b = f.as_integer_ratio()
    else:
        a,b = f.numerator,f.denominator
    if abs( a / b - x ) < tol:
        return a,b
    else:
        return None

def prettyFmtValue(x):
    """Recognises common fractions in truncated values like '0.33333333' or
    '0.4166667' and formats them as '1/3' and '5/12'. This not only saves space,
    but can also reinject accuracy in atom positions (ncrystal_verifyatompos can
    be used to check this, but it seems to be a clear improvement in all cases
    tested so far). Additionally also performs optimisations for storage
    efficiency, in particular omitting leading 0s with no information value
    (e.g. "0.2341" becomes ".2341").

    """
    assert 0.0 <= x <= 1.0
    if x==0.0:
        return '0'
    if x==1.0:
        return '1'
    if x==0.5:
        return '1/2'
    stripleading0 = lambda s : ( s[1:] if (len(s) > 2 and s.startswith('0.')) else s )
    xfmt = stripleading0('%.13g'%x)#.14g leads to irreproducibility issues in
    #our tests but there are sooo many numbers ending with 3333.... or
    #666667.... that we can safely "snap" these to their correct values:
    if xfmt[0]=='.' and xfmt.endswith('33333') and len(xfmt)<18:
        xfmt += '3'*(18-len(xfmt))
    if xfmt[0]=='.' and xfmt.endswith('66667') and len(xfmt)<18:
        xfmt = xfmt[:-1]+'6'*(18-len(xfmt))+'7'
    ff = find_fraction( x, tol = 1e-14, max_denom = 40 )
    if ff is not None and ff[1]!=1:
        assert ff[1] > 0
        v = ff[0] / ff[1]
        v_dbfmt = f'{ff[0]}/{ff[1]}'
        #check if credible that the numbers are identical (must be close AND the str
        #representations must be consistent with truncation):
        expected_at_prec = stripleading0((f'%.{len(xfmt)-1}g')%v)
        if abs(v-x)<1e-7 and expected_at_prec == xfmt:
            return v_dbfmt
        if abs(v-x)<1e-14:
            return v_dbfmt
    #No fraction found:
    if xfmt.isdigit() and float( xfmt ) != x:
        #abort, we don't want 0.99999999999 to print as '1' (which would then be
        #mapped to 0 in a unit cell):
        return stripleading0('%.19g'%x)
    return xfmt

def _split_trailing_digit( s ):
    if not s or not s[-1].isdigit():
        return s,None
    n=0
    while (n+1)<len(s) and s[-(n+1)].isdigit():
        n += 1
    if n:
        return s[0:-n], int( s[-n:] )
    else:
        return s, None

def check_elem_or_isotope_marker( s ):
    """If input is of form "Al", "O16", ..., return it, otherwise return
       None. Ignores excess whitespace in input. Will also map "H2"->"D" and
       "H3"->"T".
    """
    s=' '.join(str(s).split())
    if not s:
        return None
    elem_name, isotope_val = _split_trailing_digit( s )

    #special support for D,T:
    if not isotope_val and elem_name in ('D','T'):
        return elem_name
    if elem_name == 'H' and isotope_val in (2,3):
        return 'D' if isotope_val == 2 else 'T'

    from .atomdata import isElementName

    if isElementName( elem_name) and ( isotope_val is None or ( 0 <= isotope_val < 999 ) ):
        return elem_name if not isotope_val else '%s%i'%(elem_name,isotope_val)


def _hill_sort( chemform ):
    #takes chemform like [('Al',2'),('O',3)]  and sorts in order of Hill system notation

    #Remap H2/H3 and remove duplicates:
    remap = {'H2':'D','H3':'T'}
    if any(k in remap for k,v in chemform) or len(set(k for k,v in chemform))!=len(chemform):
        d={}
        for k,v in chemform:
            k = remap.get(k,k)
            d[k] = d.setdefault(k,0) + v
        return _hill_sort( list(d.items()) )

    has_carbon = any( en=='C' for en,c in chemform )
    if not has_carbon:
        return list( sorted( chemform ) )
    def hillsortkey( e ):
        #if not has carbon, then all in alphabetical order
        #first carbon, then H/D/T, then in alphabetical order
        if e[0]=='C':
            return ( -999999, e )
        is_hydrogen = ( e[0] in ('H','D','T') or ( e[0].startswith('H') and e[0][1:].isdigit() ) )
        return ( -1, e ) if is_hydrogen else (0, e )
    return list(sorted(chemform,key=hillsortkey))

def _gcd( *vals ):
    import math
    if len(vals)<2:
        return vals[0] if vals else None
    gcd = math.gcd(vals[0],vals[1])
    for e in vals[2:]:
        gcd = math.gcd( gcd, e )
    return gcd

def format_chemform( chemform, *, allow_rescaling = True ):
    #takes chemform like [('Al',2'),('O',3)] and returns nicely formatted string
    #"Al2O3", with no duplicated element names and sorted according to the Hill
    #system of notation. Additionally, integral counts are divided by greatest
    #common divisor, and an attempt is made to scale up non-integral counts
    #where it makes sense (e.g. Al0.5O1.5 becomes Al2O6).
    if len(chemform)==1:
        return str(chemform[0][0])
    cf = _hill_sort(chemform)
    is_near_int = lambda v : abs(int(v)-v)<1e-15
    get_non_ints = lambda _l : [ v for k,v in _l if not is_near_int(v) ]

    non_ints = get_non_ints( cf )
    _find_frac = lambda x : find_fraction( x, max_denom = 100 )
    non_int_ffractions = [ _find_frac(x) for x in non_ints ]
    denoms = [ ff[1] for ff in non_int_ffractions if ff is not None ]
    cf_alt = None
    if allow_rescaling and denoms:
        import math
        factor = 1
        while denoms:
            _d = denoms.pop(0)
            _g = math.gcd( factor, _d )
            factor *= ( _d // _g )
        _cf = []
        for k,v in cf:
            if is_near_int(v):
                _cf.append( (k,v*factor ) )
            else:
                ff = _find_frac( v )
                if ff is None:
                    _cf.append( (k,v*factor ) )
                else:
                    assert factor % ff[1] == 0
                    _cf.append( (k,ff[0] * ( factor // ff[1] ) ) )
        if max( v for k,v in _cf ) < 100:
            cf_alt = _cf

    #find and apply gcd of all integers:
    def final_format( the_cf ):
        l=[ int(v) for k,v in the_cf if is_near_int(v) ]
        gcd = _gcd( *l ) if ( l and allow_rescaling ) else 1
        wrapiso = lambda x : x if not x[-1].isdigit() else '{%s}'%x#nb: these curly braces are not great for filenames...
        the_cf = [ (wrapiso(en),(count//gcd if is_near_int(count) else count/gcd)) for en,count in the_cf ]
        return ''.join( (en if count==1 else '%s%g'%(en,int(count) if count==int(count) else count)) for en,count in the_cf )

    f1 = final_format(cf)
    f2 = final_format(cf_alt) if cf_alt else f1
    return f1 if len(f1) < len(f2) else f2


def _classifySG(sgno):
    assert 1<=sgno<=230
    l=[(195,'cubic'),(168,'hexagonal'),(143,'trigonal'),
       (75,'tetragonal'),(16,'orthorombic'),(3,'monoclinic'),(1,'triclinic')]
    for thr,nme in l:
        if sgno>=thr:
            return nme
    assert False

#colors inspired by http://www.mulinblog.com/a-color-palette-optimized-for-data-visualization/
#supposedly from Stephen Few's book, "Show Me the Numbers":
_palette_Few = dict(red = "#F15854",
                    blue="#5DA5DA",
                    orange="#FAA43A",
                    green="#60BD68",
                    brown="#B2912F",
                    purple="#B276B2",
                    yellow="#DECF3F",
                    pink="#F17CB0",
                    gray="#4D4D4D")

def _grid_is_linspace( grid, tol = 1e-6 ):
    if len(grid)<=2:
        return len(grid)==2
    from ._numpy import _ensure_numpy, _np
    _ensure_numpy()
    g = _np.asfarray(grid)
    bws = g[1:] - g[:-1]
    bmin, bmax = bws.min(), bws.max()
    if ( not ( bmax >= bmin > 0.0 ) ) or _np.isinf(bmin) or _np.isinf(bmax):
        return False
    return ( bmax - bmin ) < tol * ( bmax + bmin )

def extract_path( s ):
    """Returns pathlib.Path path object from argument if it is a path, otherwise
    None. Strings with no newlines will be assumed to be paths."""
    import pathlib
    if hasattr( s, '__fspath__' ):
        return pathlib.Path(s)
    if isinstance( s, str ) and not '\n' in s:
        return pathlib.Path(s)

def download_url( url, decode_as_utf8_str = True, wrap_exception = True ):
    import urllib.request
    import urllib.error
    try:
        req = urllib.request.Request(url)
        with urllib.request.urlopen(req) as response:
            data = response.read()
    except urllib.error.URLError as e:
        if wrap_exception:
            from .exceptions import NCException
            raise NCException(f'Error downloading url "{url}": {e}')
        else:
            raise e
    if decode_as_utf8_str:
        try:
            data = data.decode('utf8')
        except UnicodeDecodeError as e:
            if wrap_exception:
                from .exceptions import NCException
                raise NCException(f'Error decoding url to utf8 data "{url}": {e}')
            else:
                raise e
    return data

def _decodeflt(s):
    import numbers
    if s is None:
        return None
    if isinstance(s,numbers.Real):
        return float(s)
    try:
        x = float( s )
    except (TypeError,ValueError):
        return None
    return x



"""

Various constants and neutron energy / wavelength / wavenumber conversions
methods like for instance wl2ekin and ekin2wl. Note that all conversion
functions can be used with numpy arrays in addition to scalar numbers.

"""

################################################################################
##                                                                            ##
##  This file is part of NCrystal (see https://mctools.github.io/ncrystal/)   ##
##                                                                            ##
##  Copyright 2015-2023 NCrystal developers                                   ##
##                                                                            ##
##  Licensed under the Apache License, Version 2.0 (the "License");           ##
##  you may not use this file except in compliance with the License.          ##
##  You may obtain a copy of the License at                                   ##
##                                                                            ##
##      http://www.apache.org/licenses/LICENSE-2.0                            ##
##                                                                            ##
##  Unless required by applicable law or agreed to in writing, software       ##
##  distributed under the License is distributed on an "AS IS" BASIS,         ##
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  ##
##  See the License for the specific language governing permissions and       ##
##  limitations under the License.                                            ##
##                                                                            ##
################################################################################

#some constants (NB: Copied here from NCMath.hh - must keep synchronized!!):

constant_c  = 299792458e10#  speed of light in Aa/s
constant_dalton2kg =  1.660539040e-27#  amu to kg
constant_dalton2eVc2 =  931494095.17#  amu to eV/c^2
constant_avogadro = 6.022140857e23#  mol^-1
constant_boltzmann = 8.6173303e-5#  eV/K
const_neutron_mass_amu = 1.00866491588#  [amu]
constant_planck = 4.135667662e-15 # [eV*s]

kPi        = 3.1415926535897932384626433832795028841971694
k2Pi       = 6.2831853071795864769252867665590057683943388
k4Pidiv100 = 0.125663706143591729538505735331180115367886776
k4PiSq     = 39.4784176043574344753379639995046045412547976
kInf       = float('inf')
from ._numpy import _ensure_numpy, _np

_as_np_array = ( lambda x : _np.asfarray(x,dtype=float) ) if _np else ( lambda *a,**kwargs : _ensure_numpy() )
def wlsq2ekin( wlsq ):
    """Neutron wavelength squared (angstrom^2) to energy (eV)"""
    if hasattr(wlsq,'__len__'):
        x = _as_np_array( wlsq )
        wlsqnonzero = ( x != 0.0 )
        wlsqinv = 1.0 / _np.where( wlsqnonzero, x, 1.0)#fallback 1.0 wont be used
        return _const_wlsqekin * _np.where( wlsqnonzero, wlsqinv, kInf )
    else:
        return _const_wlsqekin / wlsq if wlsq else kInf

def wl2ekin( wl ):
    """Neutron wavelength (angstrom) to energy (eV)"""
    if hasattr(wl,'__len__'):
        x = _as_np_array( wl )
        return wlsq2ekin( x*x )
    else:
        return wlsq2ekin( wl*wl )

def ekin2wl( ekin ):
    """Neutron energy (eV) to wavelength (angstrom)"""
    if hasattr(ekin,'__len__'):
        x = _as_np_array( ekin )
        ekinnonzero = x != 0.0
        ekininv = 1.0 / _np.where( ekinnonzero, x, 1.0)#fallback 1.0 wont be used
        return _np.sqrt( _const_wlsqekin * _np.where( ekinnonzero, ekininv, kInf ) )
    else:
        from math import sqrt as _math_sqrt
        return _math_sqrt( _const_wlsqekin / ekin ) if ekin else kInf

def ekin2wlsq( ekin ):
    """Neutron energy (eV) to wavelength squared (angstrom^2)"""
    if hasattr(ekin,'__len__'):
        x = _as_np_array( ekin )
        ekinnonzero = x != 0.0
        ekininv = 1.0 / _np.where( ekinnonzero, x, 1.0)#fallback 1.0 wont be used
        return _const_wlsqekin * _np.where( ekinnonzero, ekininv, kInf )
    else:
        return ( _const_wlsqekin / ekin ) if ekin else kInf

def ekin2wlsqinv( ekin ):
    """Neutron energy (eV) to inverse wavelength squared (1/angstrom^2)"""
    return ekin * _const_inv_wlsqekin#constant is 1/_const_wlsqekin

_const_wlsqekin     = 0.081804209605330899    # ekin = _const_wlsqekin /wl^2
_const_inv_wlsqekin = 12.22430978582345950656 # 1 / _const_wlsqekin
_const_ekin2ksq_factor = k4PiSq * _const_inv_wlsqekin
_const_ksq2ekin_factor = 1.0 / _const_ekin2ksq_factor

def ekin2ksq( ekin ):
    """Neutron energy (eV) to wavenumber squared, (k^2, in units of 1/angstrom^2)"""
    return _const_ekin2ksq_factor * ekin

def ekin2k( ekin ):
    """Neutron energy (eV) to wavenumber, (k, in units of 1/angstrom)"""
    if hasattr(ekin,'__len__'):
        _ = _const_ekin2ksq_factor * _as_np_array( ekin )
        return _np.sqrt( _ )
    else:
        from math import sqrt
        return sqrt( _const_ekin2ksq_factor * ekin )

def ksq2ekin( ksq ):
    """Wavenumber squared, (k^2, in units of 1/angstrom^2) to neutron energy (eV)"""
    return _const_ksq2ekin_factor * ksq

def wl2k( wl ):
    """Neutron wavelength (angstrom) to wavenumber, (k, in units of 1/angstrom)"""
    if hasattr(wl,'__len__'):
        x = _as_np_array( wl )
        wlnonzero = x != 0.0
        wlinv = 1.0 / _np.where( wlnonzero, x, 1.0)#fallback 1.0 wont be used
        return k2Pi * _np.where( wlnonzero, wlinv, kInf )
    else:
        return k2Pi / wl if wl else kInf

def wl2ksq( wl ):
    """Neutron wavelength (angstrom) to wavenumber squared, (k^2, in units of 1/angstrom^2)"""
    return ( wl2k(wl) )**2

def k2wl( k ):
    """Wavenumber, (k, in units of 1/angstrom) to neutron wavelength (angstrom)"""
    return wl2k( k )#using that k2wl = wl2k (both are f(x)=2pi/x)

def k2ekin( k ):
    """Wavenumber, (k, in units of 1/angstrom) to neutron energy (eV)"""
    return ksq2ekin( k * k )



"""

Internal implementation details for utilities in ncmat.py

"""

################################################################################
##                                                                            ##
##  This file is part of NCrystal (see https://mctools.github.io/ncrystal/)   ##
##                                                                            ##
##  Copyright 2015-2023 NCrystal developers                                   ##
##                                                                            ##
##  Licensed under the Apache License, Version 2.0 (the "License");           ##
##  you may not use this file except in compliance with the License.          ##
##  You may obtain a copy of the License at                                   ##
##                                                                            ##
##      http://www.apache.org/licenses/LICENSE-2.0                            ##
##                                                                            ##
##  Unless required by applicable law or agreed to in writing, software       ##
##  distributed under the License is distributed on an "AS IS" BASIS,         ##
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  ##
##  See the License for the specific language governing permissions and       ##
##  limitations under the License.                                            ##
##                                                                            ##
################################################################################

__all__ = []

def matsrc_detect_fmt( class_MaterialSource, data ):
    if isinstance(data,class_MaterialSource):
        return 'NCrystal.MaterialSource'
    from . import core as _nc_core
    for c in ('Info','Scatter','Absorption','LoadedMaterial','TextData'):
        if isinstance(data,getattr(_nc_core,c)):
            return f'NCrystal.{c}'
    if hasattr(data,'__fspath__'):
        return 'filepath'
    if isinstance(data,bytes) or isinstance(data,str):
        compat_str = (lambda s : s.encode()) if isinstance(data,bytes) else (lambda s : s)
        if compat_str('\n') in data and data.startswith(compat_str('NCMAT')):
            return 'ncmat'
        else:
            return 'cfgstr'
    if hasattr(data,'load') and hasattr(data,'create_ncmat'):
        from .ncmat import NCMATComposer as _composer
        if isinstance(data,_composer):
            return 'NCrystal.NCMATComposer'

def matsrc_allfmts( as_str = False ):
    _= ('cfgstr','filepath','ncmat','NCrystal.TextData','NCrystal.NCMATComposer',
        'NCrystal.Info','NCrystal.Scatter','NCrystal.Absorption',
        'NCrystal.LoadedMaterial','NCrystal.MaterialSource')
    return _ if not as_str else '"%s"'%('", "'.join(_))

def matsrc_initfmt( data, fmt, cfg_params ):
    from .exceptions import NCBadInput
    cfg_params = (cfg_params or '').strip() or None
    from . import core as _nc_core
    to_str = lambda s : ( s.decode() if hasattr(s,'decode') else s )
    def cfgcombine( cfgstr, extra_params ):
        l = list( e.strip() for e in (to_str(cfgstr),to_str(extra_params)) if (e or '').strip() )
        return ';'.join(l) if l else None
    if fmt == 'NCrystal.MaterialSource':
        if not cfg_params:
            return data
        wrapped_loadfct = data.get('loadfct')
        if not wrapped_loadfct:
            raise NCBadInput('Can not add cfg_params to MaterialSource'
                             ' which is of "preloaded" type.')
        def loadfct( extra_cfg, doInfo, doScatter, doAbsorption ):
            return wrapped_loadfct( cfgcombine(cfg_params,extra_cfg), doInfo, doScatter, doAbsorption )
        descr = data.get('description')
        plotlabel = data.get('plotlabel')
        if plotlabel:
            plotlabel+=f'+ ";{cfg_params}"'
        return dict( description = f'{descr} + ";{cfg_params}"',
                     plotlabel = plotlabel,
                     loadfct = loadfct )
    if fmt == 'filepath':
        import pathlib
        p = pathlib.Path(data)
        if not p.exists():
           raise NCBadInput(f'Missing file: {p}')
        try:
            _textdata = p.read_text()
        except OSError as e:
            raise NCBadInput(f'Could not read data from file: {p}')
        except UnicodeDecodeError as e:
            raise NCBadInput(f'Problems interpreting data in file as text: {p}')
        if _textdata.startswith('NCMAT'):
            _dtype = 'ncmat'
        else:
            _ = p.suffix
            if _ and _.startswith('.'):
                _ = _[1:]
            _dtype = _ or None
        descr = p.name
        if cfg_params:
            descr += ';' + cfg_params
        def loadfct( extra_cfg, doInfo, doScatter, doAbsorption ):
            return _nc_core.directMultiCreate( _textdata,
                                               cfg_params = cfgcombine(cfg_params,extra_cfg),
                                               dtype = _dtype,
                                               doInfo = doInfo,
                                               doScatter = doScatter,
                                               doAbsorption = doAbsorption )
        return dict( description = descr, loadfct = loadfct )
    if fmt=='cfgstr':
        cfgstr = to_str(data)
        if cfg_params:
            cfgstr += ';'+to_str(cfg_params)
        from .cfgstr import normaliseCfg
        cfgstr = normaliseCfg( cfgstr )
        def loadfct( extra_cfg, doInfo, doScatter, doAbsorption ):
            cs = cfgcombine(cfgstr,extra_cfg)
            i = _nc_core.createInfo(cs) if doInfo else None
            s = _nc_core.createScatter(cs) if doScatter else None
            a = _nc_core.createAbsorption(cs) if doAbsorption else None
            return _nc_core.LoadedMaterial.fromExistingObjects(info=i,scatter=s,absorption=a)
        return dict( description = cfgstr, loadfct = loadfct )
    if fmt=='NCrystal.TextData':
        descr = data.dataSourceName or 'Anonymous TextData'
        if cfg_params:
            descr += ';' + cfg_params
        def loadfct( extra_cfg, doInfo, doScatter, doAbsorption ):
            return _nc_core.directMultiCreate( data,
                                               cfg_params = cfgcombine(cfg_params,extra_cfg),
                                               doInfo = doInfo,
                                               doScatter = doScatter,
                                               doAbsorption = doAbsorption )
        return dict( description = descr, loadfct = loadfct )
    if fmt=='ncmat':
        data = to_str(data)
        def loadfct( extra_cfg, doInfo, doScatter, doAbsorption ):
            return _nc_core.directMultiCreate( data,
                                               dtype = 'ncmat',
                                               cfg_params = cfgcombine(cfg_params,extra_cfg),
                                               doInfo = doInfo,
                                               doScatter = doScatter,
                                               doAbsorption = doAbsorption )
        return dict( description = 'Anonymous NCMAT data', loadfct = loadfct )
    if fmt=='NCrystal.NCMATComposer':
        txtdata = data.create_ncmat()# NCMATComposer is not immutable so we simply
                                     # capture the state by invoking create_ncmat
                                     # already.
        plotlabel = data.plotlabel
        def loadfct( extra_cfg, doInfo, doScatter, doAbsorption ):
            return _nc_core.directMultiCreate( txtdata,
                                               dtype = 'ncmat',
                                               cfg_params = cfgcombine(cfg_params,extra_cfg),
                                               doInfo = doInfo,
                                               doScatter = doScatter,
                                               doAbsorption = doAbsorption )
        return dict( description = 'NCMATComposer object', loadfct = loadfct, plotlabel = plotlabel )
    if fmt=='NCrystal.LoadedMaterial':
        return dict( description = 'NCrystal.LoadedMaterial object',
                     preloaded = data )
    if fmt=='NCrystal.Info':
        return dict( description = 'NCrystal.Info object',
                     preloaded = _nc_core.LoadedMaterial.fromExistingObjects( info = data ) )
    if fmt=='NCrystal.Scatter':
        return dict( description = 'NCrystal.Scatter object',
                     preloaded = _nc_core.LoadedMaterial.fromExistingObjects( scatter = data ) )
    if fmt=='NCrystal.Absorption':
        return dict( description = 'NCrystal.Absorption object',
                     preloaded = _nc_core.LoadedMaterial.fromExistingObjects( absorption = data ) )
    raise _nc_core.NCLogicError(f'unhandled fmt option: {fmt}')

def matsrc_init( class_MaterialSource, matsrc_extract_d_fct, data, fmt, cfg_params ):
    from .exceptions import NCBadInput
    assert fmt is None or isinstance(fmt,str)
    assert data is not None
    assert cfg_params is None or isinstance(cfg_params,str)
    if not fmt:
        fmt = matsrc_detect_fmt( class_MaterialSource, data )
        if not fmt:
            raise NCBadInput('Could not detect format of MaterialSource.'
                             ' If this is merely a detection problem, set'
                             ' the fmt parameter to an accepted value (one of:'
                             ' %s)'%matsrc_allfmts(as_str=True))
    if not fmt in matsrc_allfmts():
        raise NCBadInput( f'Unknown format "{fmt}" (must be one of:'
                          ' %s)'%matsrc_allfmts(as_str=True) )
    d = matsrc_initfmt( matsrc_extract_d_fct(data) if fmt=='NCrystal.MaterialSource' else data, fmt, cfg_params )
    assert len(d) in (2,3)
    assert 'description' in d
    assert int('preloaded' in d) + int('loadfct' in d) == 1
    if 'preloaded' in d and (cfg_params or '').strip():
        raise NCBadInput('MaterialSource objects that are wrapping preloaded physics objects '
                         'can not be initialised with additional cfg parameters.')
    return d

def anytextdata_init( data, *, is_path, name ):
    if isinstance(name,bytes):
        name = name.decode()
    assert name is None or isinstance(name,str)
    res_keepalive = None
    if hasattr(data,'rawData') and hasattr(data,'dataSourceName'):
        #TextData:
        res_keepalive = data
        res_name = name or data.dataSourceName
        res_content = data.rawData
        return res_name, res_content, res_keepalive
    if hasattr( data, '__fspath__' ):
        import pathlib
        p = pathlib.Path(data)
        res_name = name or p.name
        res_content = p.read_text()
        return res_name, res_content, res_keepalive
    if isinstance(data,bytes):
        data = data.decode()
    if not isinstance(data,str):
        from .exceptions import NCBadInput
        raise NCBadInput('Invalid text data / text file data (got type %s)'%type(data))
    is_path = is_path if ( not is_path is None ) else ( not '\n' in data )
    if is_path:
        import pathlib
        p = pathlib.Path ( data )
        res_name = name or p.name
        res_content = p.read_text()
    else:
        res_content = data
        res_name = name#Ok if None
    return res_name or None, res_content, res_keepalive

def _anyvdos_detect_fmt( class_AnyVDOS, data ):
    if isinstance(data,class_AnyVDOS):
        return 'NCrystal.AnyVDOS'
    from .core import Info
    if isinstance( data, Info.DI_VDOS ):
        return 'NCrystal.Info.DI_VDOS'
    if isinstance( data, Info.DI_VDOSDebye ):
        return 'NCrystal.Info.DI_VDOSDebye'
    import numbers
    if isinstance(data, numbers.Real):
        return 'debyetemperature'
    if hasattr(data,'__len__') and len(data)==2:
        eg,dens = data
        if hasattr(eg,'__len__') and len(eg) >= 2 and hasattr(dens,'__len__') and len(dens)>=2 and ( len(eg)==len(dens) or len(eg)==2 ):
            return 'arrays'

def _anyvdos_preinit( data, fmt ):
    allfmts = ('NCrystal.Info.DI_VDOS','NCrystal.Info.DI_VDOSDebye','debyetemperature','arrays','NCrystal.AnyVDOS')
    from ._numpy import _ensure_numpy, _np
    _ensure_numpy()

    if fmt =='NCrystal.Info.DI_VDOS':
        return dict( dos_orig = ( data.vdosOrigEgrid(), data.vdosOrigDensity() ),
                     dos = data.vdosData() )
    if fmt =='NCrystal.Info.DI_VDOSDebye':
        return dict( dos = data.vdosData(),
                     debyetemp = data.debyeTemperature() )
    if fmt =='debyetemperature':
        debyetemp = float(data)
        from .vdos import createVDOSDebye
        return dict( dos = createVDOSDebye( debyetemp ),
                     debyetemp = debyetemp )
    if fmt =='arrays':
        eg, dens = data
        dens = _np.asfarray( dens, dtype = float )
        if len(eg)==2 and len(dens)>2:
            eg = (float(eg[0]),float(eg[1]))
        else:
            eg = _np.asfarray( eg, dtype = float )
        return dict( dos = ( eg, dens ) )
    assert not fmt == 'NCrystal.AnyVDOS'#should have been handled in calling code
    if not fmt in allfmts:
        from .exceptions import NCBadInput
        s='"%s"'%('", "'.join(allfmts))
        if fmt is None:
            raise NCBadInput('Could not detect fmt of VDOS specification, consider specifying it explictly (must be one of %s)'%(s))
        raise NCBadInput('Invalid VDOS fmt "%s". Possible options are %s'%(fmt,s))
    assert False, "unhandled fmt case: %s"%fmt

def _anyvdos_initfmt( data, fmt ):
    p = _anyvdos_preinit( data, fmt )
    #find derived quantities:
    d = {}
    from ._numpy import _ensure_numpy, _np_linspace, _np
    _ensure_numpy()
    def _needsexpand( egrid, dos ):
        return len(egrid)==2 and len(dos)>2
    def _expandegrid( egrid, dos ):
        if _needsexpand(egrid,dos):
            return _np_linspace( egrid[0], egrid[-1], len(dos) )
        else:
            return _np.asfarray(egrid,dtype=float)
    def integrate( expanded_egrid, dos ):
        x,y = expanded_egrid, dos
        if not x[0] > 0.0:
            #simply ignore points <= 0.0
            i = _np.argmax(x>0.0)
            x,y = x[i:],y[i:]
        parabola_contrib = ( x[0] * y[0] ) / 3.0#integral of parabola through (0,0) and (x[0],y[0]) over [0,x[0]]
        return parabola_contrib + _np.trapz(x=x,y=y)
    for k in ('dos','dos_orig'):
        if not k in p:
            continue
        eg,dos = p[k]
        egexpand = _expandegrid(eg,dos) if _needsexpand(eg,dos) else None
        d[k+'_expandedegrid'] = egexpand
        d[k+'_integral'] = integrate( ( eg if egexpand is None else egexpand), dos )

    #combine and return:
    import copy
    res = {'derived':d}
    for k,v in p.items():
        res[k] = copy.deepcopy(v)
    return res

def _anyvdos_init( class_AnyVDOS, anyvdos_extract_d, data, fmt, label ):
    if label is None and hasattr(data,'_plotlabel'):
        from .core import Info
        if isinstance(data,Info.DynamicInfo):
            label = data._plotlabel()
    fmt = _anyvdos_detect_fmt(class_AnyVDOS,data) if fmt is None else fmt
    if fmt == 'NCrystal.AnyVDOS':
        d = anyvdos_extract_d( data )
        if label is None or d.get('label')==label:
            return d
        else:
            import copy
            d = dict( (k,copy.deepcopy(v)) for k,v in d.items() if k != 'label' )
    else:
        d = _anyvdos_initfmt( data, fmt )
    d['label'] = label or 'anonymous VDOS curve from "%s"'%fmt
    if d.get('debyetemp'):
        _='TDebye=%gK'%d['debyetemp']
        if not d['label'].endswith(')'):
            d['label'] += f' ({_})'
        else:
            d['label'] = d['label'][:-1]+f', {_})'
    from types import MappingProxyType#read-only dict
    d['derived'] = MappingProxyType(d['derived'])
    return MappingProxyType( d )

def detect_scatcomps( standard_comp_types, matsrc ):
    if matsrc.is_preloaded:
        from .exceptions import NCBadInput
        raise NCBadInput('detect_scattering_components can not be used'
                         ' with preloaded material sources')
    res=[]
    for ct in standard_comp_types:
        _ = matsrc.load( extra_cfg_params = f'vdoslux=0;comp={ct}', doInfo = False, doAbsorption = False )
        if not _.scatter.isNull():
            res.append(ct)
    return res




################################################################################
##                                                                            ##
##  This file is part of NCrystal (see https://mctools.github.io/ncrystal/)   ##
##                                                                            ##
##  Copyright 2015-2023 NCrystal developers                                   ##
##                                                                            ##
##  Licensed under the Apache License, Version 2.0 (the "License");           ##
##  you may not use this file except in compliance with the License.          ##
##  You may obtain a copy of the License at                                   ##
##                                                                            ##
##      http://www.apache.org/licenses/LICENSE-2.0                            ##
##                                                                            ##
##  Unless required by applicable law or agreed to in writing, software       ##
##  distributed under the License is distributed on an "AS IS" BASIS,         ##
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  ##
##  See the License for the specific language governing permissions and       ##
##  limitations under the License.                                            ##
##                                                                            ##
################################################################################

"""

Utility functions related to plotting.

"""

def plot_xsect( material, *, mode='wl', do_show = True, do_newfig = True,
                npts=5000, xmin = None, xmax = None, ymin=None, ymax=None,
                logx = None, logy=None,
                scatter_breakdown = True, xsmode='peratom',
                show_absorption = True, show_scattering = True,
                do_legend = True, do_grid = True, color = None,
                labelfct = None, only_total = False, extra_cfg=None,
                title = None ):
    """Quick plot (with matplotlib) showing the cross sections produced by
    current material. The mode parameter can be either 'wl' or 'ekin', and
    logx/logy can be set to None or a boolean to control whether the
    corresponding axis should be logarithmic (None indicates False when
    mode=='wl' and True when mode=='ekin'). Set scatter_breakdown to False
    to show the total scattering contribution rather than a breakdown into
    components (only affects materials that are not preloaded).

    The extra_cfg parameter can be used to append cfg parameters
    (e.g. "comp=bragg,incoh_elas;temp=50K").

    Use the plot_xsects (note the trailing "s") function instead to compare
    cross sections from multiple materials.

    """
    from . import misc as nc_misc
    from .exceptions import NCBadInput
    matsrc = nc_misc.MaterialSource( material, cfg_params = extra_cfg )
    assert mode in ('wl','ekin')
    assert logx is None or logx in (True,False)
    assert xsmode in ('macroscopic','peratom')
    loadedmat = matsrc.load()
    if not loadedmat.scatter and not loadedmat.absorption:
        raise NCBadInput('Can not produce plots for material source which contains no physics processes')
    do_absn = show_absorption and bool(loadedmat.absorption)
    do_scat = show_scattering and bool(loadedmat.scatter)
    labelfct = labelfct or ( lambda x : x )
    plt = _import_matplotlib_plt()
    if do_newfig:
        plt.figure()
    from ._numpy import _np_linspace, _np_geomspace, _np, _ensure_numpy
    _ensure_numpy()
    from .constants import wl2ekin
    from ._common import _palette_Few as _palette
    wlmax = _estimate_longest_interesting_wavelength( loadedmat.info )
    if mode=='wl':
        xmin = 0.0 if xmin is None else xmin
        xmax = wlmax if xmax is None else xmax
        logxydefault = False
    else:
        xmin = ( wl2ekin(wlmax)*1e-3 ) if ( xmin is None ) else ( xmin or 1e-7 )
        xmax = xmax or xmin*1e5
        logxydefault = True
    assert xmax > xmin
    logx = logxydefault if logx is None else bool(logx)
    logy = logxydefault if logy is None else bool(logy)
    x = (_np_geomspace if logx else _np_linspace)( xmin, xmax, npts )
    xsectargs = dict(wl=x) if mode=='wl' else dict(ekin=x)
    if mode=='macroscopic' and not loadedmat.info:
        raise NCBadInput('Can not produce macroscopic cross section'
                         ' plots for material source which contains'
                         ' no NCrystal.Info object (and therefore no material density).')
    xsectfactor = ( loadedmat.info.factor_macroscopic_xs ) if ( xsmode=='macroscopic' and loadedmat.info ) else 1.0
    do_scatter_breakdown = do_scat and scatter_breakdown and not matsrc.is_preloaded and not loadedmat.scatter.isNull()
    xs_s = xsectfactor * loadedmat.scatter.xsect(**xsectargs) if do_scat else None
    xs_a = xsectfactor * loadedmat.absorption.xsect(**xsectargs) if do_absn else None
    xs_tot = xs_s + xs_a if ( do_scat and do_absn ) else ( xs_s if do_scat else xs_a )
    nullxs = not ( xs_tot.max() > 0.0 )
    if nullxs and logy:
        from ._common import warn
        warn('Could not set log scale since curves are 0.0 everywhere')
        logy=False

    plotcalls = []
    def add_plot(x,y,*a,**kw):
        if y is not None:
            #if logy:
            #    nonzeroy = ( y > 0.0 )
            #    x,y = x[nonzeroy],y[nonzeroy]
            plotcalls.append( (x,y,a,kw) )
    if do_scatter_breakdown:
        lblmap = { 'coh_elas':'Coherent elastic',
                   'incoh_elas':'Incoherent elastic',
                   'inelas':'Inelastic',
                   'sans':'SANS'}
        colmap = { 'coh_elas':'blue',
                   'incoh_elas':'yellow',
                   'inelas':'green',
                   'sans':'pink'}#nb unused: brown, gray (but brown looks similar to yellow)

        assert set(lblmap.keys())==set(nc_misc.standard_comp_types),"update lblmap"
        assert set(colmap.keys())==set(nc_misc.standard_comp_types),"update colmap"
        for comp in nc_misc.detect_scattering_components( matsrc ):
            _xs_comp = xsectfactor * matsrc.load(f'comp={comp}',doInfo=False,doAbsorption=False).scatter.xsect(**xsectargs)
            add_plot(x,_xs_comp,label=labelfct(lblmap[comp]),
                     color = color or _palette[colmap[comp]] )
    else:
        add_plot(x,xs_s,label=labelfct('Scattering'),
                 color = color or _palette['blue'])
    add_plot(x,xs_a,label=labelfct('Absorption'),color = color or _palette['purple'])

    if only_total:
        plotcalls=[]
    if only_total or len(plotcalls) > 1:
        totlabel = 'Total' if ( do_absn and do_scat ) else ( 'Total scattering' if do_scat else 'Total absorption' )
        add_plot(x,xs_tot,label=labelfct(totlabel),
                 color=color or _palette['red'])
    for x,y,a,kw in plotcalls:
        plt.plot(x,y,*a,**kw)

    plt.xlabel('Neutron wavelength (angstrom)' if mode=='wl' else 'Neutron energy (eV)')
    plt.ylabel('Macroscopic cross section (1/cm)' if xsmode=='macroscopic' else 'Cross section per atom (barn)')

    auto_ymin = ( None if logy else 0.0)
    auto_ymax = ( 1.0 if nullxs else None )
    ymin = ( auto_ymin if ymin is None else ymin )
    ymax = ( auto_ymax if ymax is None else ymax )
    if ymin is not None or ymax is not None:
        if ymin is not None and ymax is not None and not ymax>ymin:
            from ._common import warn
            _fmt = lambda x : ( 'auto' if x is None else x )
            warn('ymin/ymax parameters would lead to a plot range of'
                 f' [{ymin},{ymax}]. Reverting to [{_fmt(auto_ymin)},{_fmt(auto_ymax)}].')
            ymin,ymax = auto_ymin,auto_ymax
        plt.ylim(ymin,ymax)
    plt.xlim(xmin,xmax)
    if title is not False:
        plt.title(title or matsrc.plotlabel)
    _plt_final(do_grid,do_legend,do_show,logy=logy,logx=logx)

def plot_xsects( *materials, **plotkwargs ):
    """Compares cross sections of multiple materials. Accepts similar keywords
    as the plot_xsect function."""
    from .misc import MaterialSource
    mats = [ MaterialSource(m) for m in materials ]
    do={}
    for e in ['do_newfig','do_legend','do_grid','do_show','logy','logx']:
        do[e] = plotkwargs.get(e,None if e in ('logy','logx') else True)
        plotkwargs[e] = False
    title = plotkwargs.get('title')
    plotkwargs['title'] = False
    plotkwargs['scatter_breakdown']=False
    plotkwargs['only_total'] = True
    plt = _import_matplotlib_plt()
    if do['do_newfig']:
        plt.figure()
    col_ordered = _get_col_ordered()
    for i,m in enumerate(mats):
        plotkwargs['color'] = col_ordered[i%len(col_ordered)]
        plotkwargs['labelfct'] = lambda l : m.plotlabel
        if i+1 == len(mats):
            #the last:
            plotkwargs['logy'] = do['logy']
            plotkwargs['logx'] = do['logx']
        plot_xsect(m,**plotkwargs)
    if title:
        plt.title(title)
    _plt_final(do['do_grid'],do['do_legend'],do['do_show'])

def plot_vdos( *vdos, unit='meV',
               show_orig_data = False,
               show_normalised = True,
               do_show = True, do_newfig = True,
               npts_parabola=5000, logy=False,
               do_legend = True, do_grid = True, color = None,
               labelfct = None ):
    """
    Quick plot (with matplotlib) showing the requested VDOS curve(s).
    """
    from . import misc as nc_misc
    from . import vdos as nc_vdos
    #from ._common import _palette_Few as _palette
    from .exceptions import NCBadInput
    from ._numpy import _np_linspace,_ensure_numpy, _np
    unit_name, unit_value = nc_vdos._parsevdosunit( unit )

    #from ._numpy import _np_linspace, _np_geomspace, _np, _ensure_numpy
    #_ensure_numpy()
    #from .constants import wl2ekin

    vdoslist = [ ( nc_misc.AnyVDOS(v),False) for v in vdos ]
    if show_orig_data:
        l=[]
        for v,_ in vdoslist:
            l.append( (v, False ) )
            if v.has_orig:
                l.append( (v, True ) )
        vdoslist = l

    plt = _import_matplotlib_plt()
    if do_newfig:
        plt.figure()

    col_ordered = _get_col_ordered()

    for ivdos,(vdos,use_orig) in enumerate(vdoslist):
        lbl = vdos.label or 'VDOS#%i'%(ivdos+1)
        if use_orig:
            lbl += ' (original)'
        if labelfct:
            lbl = labelfct( lbl )
        x = vdos.egrid( orig = use_orig ) / unit_value
        y = vdos.dos( orig = use_orig, norm = show_normalised ) * unit_value
        color = col_ordered[ivdos%len(col_ordered)]
        if x[0] > 0.0 and npts_parabola:
            _ensure_numpy()
            xp = _np_linspace( 0.0, x[0], npts_parabola )
            plt.plot(xp, y[0] *(xp/x[0])**2, color = color, ls = ':' )
        if y[-1] > 0.0:
            _ensure_numpy()
            x = _np.append( x, [ x[-1] ] )
            y = _np.append( y, [ 0.0 ] )
        plt.plot(x,y,color = color, label = lbl, marker='.' )

    plt.xlabel(f'Frequency ({unit_name})')
    plt.ylabel('VDOS (arbitrary units)')
    _plt_final(do_grid,do_legend,do_show,logy=logy)

def plot_knl( kernel, do_newfig = True, do_show = True, do_grid = True, logz=False, phasespace_curves = None,
              clim=None, xlim = None, ylim = None):
    """Quick plot (with matplotlib) showing the requested scattering kernel in
    in S(alpha,beta) format). The kernel is assume to be a dictionary with keys
    'alpha', 'beta', and 'sab' associated with values being numpy arrays
    containing the relevant info. Optionally, keys 'egrid' and 'suggestedEmax'
    can be present, and will be used to estimate Emax - the highest intended
    neutron energy for which the table is to be used. Emax is only used to plot
    the kinematic bound of Emax.
    """
    #estimate emax:
    _egrid = kernel.get('egrid')
    _suggestedEmax = kernel.get('suggestedEmax')
    emax = None
    if _suggestedEmax and _suggestedEmax>0.0:
        emax = _suggestedEmax
    if emax is None and _egrid is not None:
        import numbers
        if isinstance(_egrid,numbers.Real):
            emax = float(_egrid)
        elif len(_egrid)==1:
            emax = _egrid[0]
        elif len(_egrid)==3:
            emax = _egrid[1]
        elif len(_egrid)>3:
            emax = _egrid[-1]

    alpha = kernel['alpha']
    beta = kernel['beta']
    sab = kernel['sab']
    assert len(alpha)>=5
    assert len(beta)>=5
    assert len(alpha)*len(beta) == len(sab)

    from ._numpy import _ensure_numpy, _np, _np_linspace
    _ensure_numpy()
    plt = _import_matplotlib_plt()
    if do_newfig:
        plt.figure()


    def plotedges(x):
        """given array [a,b,c,...,x,y,z] returns [a,(a+b)2,(b+c)/2,...,(y+z)/2,z]"""
        return _np.concatenate([[x[0]],0.5*(x[:-1]+x[1::1]),[x[-1]]])
    na,nb = len(alpha),len(beta)
    pa = plotedges(alpha)
    pb = plotedges(beta)
    x,y=_np.meshgrid(pa,pb)
    if logz:
        sab = _np.log(sab)
    sab=sab.reshape((nb,na))

    cmap = 'jet'#fallback for some special options
    sab_max = sab.max()
    quadmesh = plt.pcolormesh( x,y,sab, clim=clim,cmap=cmap )

    def _real_mpl_object( o ):
        #in case we are running under NCRYSTAL_FAKEPYPLOT=log, we are dealing
        #with a wrapped object - which we can not pass directly to matplotlib
        #classes.
        return o.the_real_inspected_object if hasattr(o,'the_real_inspected_object') else o

    plt.colorbar(_real_mpl_object(quadmesh))
    if clim:
        quadmesh.set_clim(clim)

    plt.xlim(*(xlim or (0.0,alpha[-1])))
    plt.ylim(*(ylim or (beta[0],beta[-1])))

    def alpha_limits(E_div_kT,beta):
        c = E_div_kT
        cb = c + beta
        assert cb>=0.0,f'c+beta={cb}, beta={beta}, c={c}'
        a = cb + c
        b = 2*_np.sqrt( c * cb )
        assert not _np.isinf(a).any()
        assert not _np.isinf(b).any()
        return max(0.0,a - b), a + b
    alpha_minus = _np.vectorize(lambda *a : alpha_limits(*a)[0])
    alpha_plus = _np.vectorize(lambda *a : alpha_limits(*a)[1])
    def kin_curve( E_div_kT ):
        b0, b1 = b0,b1=max(beta.min(),-E_div_kT),beta.max()
        #Smoother curve by putting more points around beta=0, less points
        #further out (actually, we could improve this alg when
        #ekin_div_kT>>1):
        if b0<1.0 and b1>1.0:
            bvals=_np.append(_np_linspace(b0,1.0,2500), _np_linspace(1.0+1e-14,b1,2500))
        else:
            bvals=_np_linspace(b0,b1,5000)
        color='yellow'
        plt.plot(alpha_minus(E_div_kT,bvals),bvals,label=f'E/kT={E_div_kT}',color=color,lw=2)
        plt.plot(alpha_plus(E_div_kT,bvals).clip(0,alpha.max()),bvals,color=color,lw=2)

    from .constants import constant_boltzmann
    kT = kernel['temperature']*constant_boltzmann

    for ekin in (phasespace_curves or []):
        kin_curve( ekin / kT )

    plt.xlabel('alpha')
    plt.ylabel('beta')
    _plt_final(do_grid=do_grid,do_legend=False,do_show=do_show)

def plot_vdos_Gn( Gn, unit = 'meV', logy = False,
                  do_newfig = True, do_legend = True,
                  do_grid = True, do_show = True ):
    """Plot Sjolander Gn functions. Each Gn must be specified as a sequence like
    (egrid,Gnvals,n[,label]), where either len(egrid)==len(Gnvals) or
    egrid=(emin,emax). The Gn argument can either be a single Gn function, or a
    sequence of them.
    """
    def decode_Gn( x ):
        if len(x) == 3:
            egrid, gnvals, n = x
            label = None
        elif len(x) == 4:
            egrid, gnvals, n, label = x
        else:
            return None
        if len(egrid)==2 and len(gnvals)>2:
            from ._numpy import _ensure_np,_np_linspace
            _ensure_np()
            egrid = _np_linspace(egrid[0],egrid[1],len(gnvals))
        return dict( egrid = egrid,
                     gnvals = gnvals,
                     n = n,
                     label = f'{label} (G{n})' if label else f'G{n}' )
    _ = decode_Gn( Gn )
    if _ is not None:
        gnlist = [ _ ]
    else:
        gnlist = [ decode_Gn(e) for e in Gn ]
    if not gnlist or any( e is None for e in gnlist ):
        from .exceptions import NCBadInput
        raise NCBadInput('Invalid specification of Gn functions.')

    from . import vdos as nc_vdos
    #from ._common import _palette_Few as _palette
    unit_name, unit_value = nc_vdos._parsevdosunit( unit )

    plt = _import_matplotlib_plt()
    if do_newfig:
        plt.figure()

    for gn in gnlist:
        plt.plot( gn['egrid']/unit_value, gn['gnvals'], label = gn['label'] )

    plt.xlabel(f'Energy ({unit_name})')
    _plt_final(do_grid,do_legend,do_show,logy=logy)

def _get_col_ordered():
    from ._common import _palette_Few
    return [_palette_Few.get(k,k) for k in
            ('blue','orange','green','red','brown',
             'purple','yellow','pink','gray','black')]

def _plt_final(do_grid,do_legend,do_show,logx=False,logy=False):
    plt=_import_matplotlib_plt()
    if logx:
        plt.semilogx()
    if logy:
        plt.semilogy()
    if do_legend:
        leg=plt.legend()
        if do_legend=='draggable':
            leg.set_draggable(True)
    if do_grid:
        plt.grid()
    if do_show:
        plt.show()

_fakepyplot_mode_cache = [None]
def _fakepyplot_mode():
    if _fakepyplot_mode_cache[0] is None:
        import os
        _ = os.environ.get('NCRYSTAL_FAKEPYPLOT')
        if not _:
            res = False
        else:
            res = 'log_and_plot' if ( _ == 'log' ) else 'log_and_block'
        _fakepyplot_mode_cache[0] = res
    return _fakepyplot_mode_cache[0]

_theplt=[None]
def _import_matplotlib_plt():
    if not _theplt[0]:
        mode = _fakepyplot_mode()
        if mode:
            from ._testimpl import _create_pyplot_inspector
            _theplt[0] = _create_pyplot_inspector( pass_calls_to_real_plt = ( mode == 'log_and_plot' ) )
        else:
            #first matplotlib alone, for a perhaps more pedagogical ImportError.
            import matplotlib
            import matplotlib.pyplot as plt
            _theplt[0]=plt
    return _theplt[0]

_thepdfpages=[None]
def _import_matplotlib_pdfpages():
    if not _thepdfpages[0]:
        def _raw_import_pdf_pages():
            try:
                from matplotlib.backends.backend_pdf import PdfPages
            except ImportError:
                raise ImportError("ERROR: Your installation of matplotlib does not have the required support for PDF output.")
            return PdfPages
        mode = _fakepyplot_mode()
        if mode!='log_and_block':
            import matplotlib
            matplotlib.use('agg')
        if mode:
            from ._testimpl import _create_pdfpages_inspector
            _thepdfpages[0] = _create_pdfpages_inspector( _raw_import_pdf_pages() if mode != 'log_and_block' else None )
        else:
            _thepdfpages[0] = _raw_import_pdf_pages()
    return _thepdfpages[0]

def _find_highest_bragg_edge( info ):
    if info.isSinglePhase():
        return info.braggthreshold
    l = [_find_highest_bragg_edge(p) for frac,p in info.phases]
    l = [e for e in l if e]
    return max(l) if l else None

def _estimate_longest_interesting_wavelength( info ):
    #longest wavelength of interest in material, for the purpose of plotting. In
    #case of bragg edges, this will be the longest bragg edge found in the
    #material.
    _ = _find_highest_bragg_edge( info ) if info else None
    return _*1.2 if _ else 15.0
